"""
Concrete :class:`~.base.TrackerJobsBase` subclass for PTP
"""

import functools
import re
from datetime import datetime

from ... import errors, jobs, utils
from ...utils.release import ReleaseType
from ..base import TrackerJobsBase
from . import metadata

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PtpTrackerJobs(TrackerJobsBase):

    @functools.cached_property
    def jobs_before_upload(self):
        return (
            # Common background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,
            self.ptp_group_id_job,
            self.description_job,
            self.trumpable_job,

            # Common interactive jobs
            self.type_job,
            self.imdb_job,
            self.source_job,
            # self.video_codec_job,
            # self.container_job,
            # self.resolution_job,
            self.scene_check_job,

            # Interactive jobs that only run if movie does not exists on PTP yet
            self.title_job,
            self.year_job,
            self.plot_job,
            self.tags_job,
            self.cover_art_job,
        )

    @property
    def isolated_jobs(self):
        """
        Sequence of job attribute names (e.g. "title_job") that were singled
        out by the user, e.g. with --only-title
        """
        if self.options.get('only_description', False):
            return self.get_job_and_dependencies(
                self.description_job,
                # `screenshots_job` is needed by `upload_screenshots_job`, but
                # `upload_screenshots_job` is a `QueueJobBase`, which doesn't
                # know anything about the job it gets input from and therefore
                # can't tells us that it needs `screenshots_job`.
                self.screenshots_job,
            )
        else:
            # Activate all jobs in jobs_before/after_upload
            return ()

    @functools.cached_property
    def type_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('type'),
            label='Type',
            precondition=self.make_precondition('type_job'),
            autodetect=self.autodetect_type,
            options=(
                ('Feature Film', 'Feature Film'),
                ('Short Film', 'Short Film'),
                ('Miniseries', 'Miniseries'),
                ('Stand-up Comedy', 'Stand-up Comedy'),
                ('Live Performance', 'Live Performance'),
                ('Movie Collection', 'Movie Collection'),
            ),
            callbacks={
                'finished': self.update_imdb_query,
            },
            **self.common_job_args(),
        )

    def autodetect_type(self, _):
        if self.release_name.type == ReleaseType.season:
            return 'Miniseries'

        # Short film if runtime 45 min or less (Rule 1.1.1)
        # NOTE: This doesn't work for VIDEO_TS releases. We could sum the
        #       duration of all .VOB files, but they notoriously lie about their
        #       duration. For now, the user can always correct the autodetected
        #       type.
        main_video = tuple(utils.video.filter_main_videos(
            tuple(utils.video.find_videos(self.content_path))
        ))[0]
        if utils.video.duration(main_video) <= 45 * 60:
            return 'Short Film'

    def update_imdb_query(self, type_job_):
        """
        Set :attr:`~.webdbs.common.Query.type` on
        :attr:`~.TrackerJobsBase.imdb_job` to :attr:`~.ReleaseType.movie` or
        :attr:`~.ReleaseType.series` depending on the output of :attr:`type_job`
        """
        assert self.type_job.is_finished
        if self.type_job.output:
            new_type = self.type_job.output[0]
            if new_type == 'Feature Film':
                _log.debug('Updating IMDb query type: %r', utils.release.ReleaseType.movie)
                self.imdb_job.query.type = utils.release.ReleaseType.movie
            elif new_type == 'Miniseries':
                _log.debug('Updating IMDb query type: %r', utils.release.ReleaseType.series)
                self.imdb_job.query.type = utils.release.ReleaseType.series

    @functools.cached_property
    def trumpable_job(self):
        return jobs.custom.CustomJob(
            name=self.get_job_name('trumpable'),
            label='Trumpable',
            precondition=self.make_precondition('trumpable_job'),
            worker=self.autodetect_trumpable,
            no_output_is_ok=True,
            catch=(
                errors.ContentError,
            ),
            **self.common_job_args(ignore_cache=True),
        )

    async def autodetect_trumpable(self, job):
        reasons = []

        if self.options.get('hardcoded_subtitles', None) in (True, False):
            # User specified "Hardcoded subtitles"
            if self.options['hardcoded_subtitles']:
                reasons.append(metadata.PtpTrumpableReason.HARDCODED_SUBTITLES)
        # Autodetection of hardcoded subtitles would go here

        if self.options.get('no_english_subtitles', None) in (True, False):
            # User specified "No English subtitles"
            if self.options['no_english_subtitles']:
                reasons.append(metadata.PtpTrumpableReason.NO_ENGLISH_SUBTITLES)
        else:
            # Autodetect "No English subtitles":
            #   - No English audio track + No English subtitle track
            #   - Also look at external subtitle files (see utils.subtitle)
            #   - Unknown language is considered English
            audio_languages = utils.video.languages(
                self.content_path,
                default='',
            ).get('Audio', ())
            _log.debug('Trumpable reasons: Audio languages: %r', audio_languages)
            if (
                    # Not a silent film
                    audio_languages
                    # No English audio track
                    and 'en' not in audio_languages
            ):
                # English subtitles are required
                subtitle_languages = {
                    subtitle.language
                    for subtitle in self.subtitles
                }
                _log.debug('Trumpable reasons: Subtitle languages: %r', subtitle_languages)
                if 'en' not in subtitle_languages:
                    # Assume unspecified language means English
                    if (
                            '' not in subtitle_languages
                            and '' not in audio_languages
                    ):
                        # Every track has a language and all of them are
                        # specifically not English
                        reasons.append(metadata.PtpTrumpableReason.NO_ENGLISH_SUBTITLES)

        _log.debug('Trumpable reasons: %r', reasons)
        return reasons

    @functools.cached_property
    def description_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            precondition=self.make_precondition('description_job'),
            prejobs=self.get_job_and_dependencies(
                self.mediainfo_job,
                self.upload_screenshots_job,
            ),
            text=self.generate_description,
            read_only=True,
            hidden=True,
            finish_on_success=True,
            **self.common_job_args(ignore_cache=True),
        )

    async def generate_description(self):
        original_release_name = (
            '[size=4]'
            + '[b]'
            + utils.fs.basename(
                utils.fs.strip_extension(self.content_path)
            )
            + '[/b]'
            + '[/size]'
        )

        # Join mediainfos + screenshots for each movie/episode into
        # sections. Each section may contain multiple mediainfos (.IFO/.VOB).
        tags = []
        subtags = []
        for video_path, info in self.mediainfos_and_screenshots.items():
            if info['mediainfo']:
                subtags.append(f'[mediainfo]{info["mediainfo"]}[/mediainfo]')
            if info['screenshot_urls']:
                subtags.append('\n'.join(
                    f'[img={url}]'
                    for url in info['screenshot_urls']
                ))
                tags.append(''.join(subtags))
                subtags.clear()

        description = (
            original_release_name
            + '\n\n'
            + '\n[hr]\n'.join(tags)
        )
        return description

    mediainfo_from_all_videos = True
    screenshots_from_all_videos = True

    @functools.cached_property
    def screenshots_count(self):
        """
        How many screenshots to make

        Use :attr:`options`\\ ``["screenshots"]`` it it exists. This value
        should be explicitly set by the user, e.g. via a CLI argument or GUI
        element.

        Use :attr:`options`\\ ``["screenshots_from_movie"]`` for uploads that
        contain a single video file.

        Use :attr:`options`\\ ``["screenshots_from_episode"]`` for uploads that
        contain multiple video files.
        """
        # CLI option, GUI widget, etc
        if self.options.get('screenshots'):
            return self.options['screenshots']

        # Get that same video files that were passed to ScreenshotsJob in
        # JobBase.screenshots_job.
        video_files = utils.torrent.filter_files(
            self.content_path,
            exclude=self.exclude_files,
        )
        # Exclude samples, extras, DVD menus, etc
        video_files = utils.video.filter_main_videos(video_files)
        if len(video_files) == 1:
            return self.options['screenshots_from_movie']
        else:
            return self.options['screenshots_from_episode']

    @property
    def imdb_id(self):
        if self.imdb_job.is_finished:
            return self.imdb_job.output[0]

    @functools.cached_property
    def ptp_group_id_job(self):
        return jobs.custom.CustomJob(
            name=self.get_job_name('ptp-group-id'),
            label='PTP Group ID',
            precondition=self.make_precondition('ptp_group_id_job'),
            prejobs=self.get_job_and_dependencies(
                self.imdb_job,
            ),
            worker=self.fetch_ptp_group_id,
            catch=(
                errors.RequestError,
            ),
            **self.common_job_args(),
        )

    async def fetch_ptp_group_id(self, _):
        assert self.imdb_job.is_finished
        group_id = await self.tracker.get_ptp_group_id_by_imdb_id(self.imdb_id)
        return '' if group_id is None else group_id

    @property
    def ptp_group_id(self):
        """
        PTP group ID if :attr:`ptp_group_id_job` is finished and group ID
        was found, `None` otherwise
        """
        if self.ptp_group_id_job.is_finished:
            if self.ptp_group_id_job.output:
                return self.ptp_group_id_job.output[0]

    def ptp_group_does_not_exist(self):
        """
        Whether no releases of :attr:`imdb_id` already exist

        :attr:`ptp_group_id_job` must be finished when this is called.

        This is used as a :attr:`~.JobBase.precondition` for jobs that are only
        needed if the server doesn't have any releases for this :attr:`imdb_id`
        yet.
        """
        assert self.ptp_group_id_job.is_finished, self.ptp_group_id_job
        if self.ptp_group_id:
            return False
        else:
            return True

    @functools.cached_property
    def title_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('title'),
            label='Title',
            precondition=self.make_precondition(
                'title_job',
                # Don't run this job if PTP already knows this movie.
                precondition=self.ptp_group_does_not_exist,
            ),
            prejobs=self.get_job_and_dependencies(
                # We need to wait for the PTP group ID to become available
                # before we can check if it exists.
                self.ptp_group_id_job,
            ),
            text=self.fetch_title,
            normalizer=self.normalize_title,
            validator=self.validate_title,
            **self.common_job_args(),
        )

    async def fetch_title(self):
        assert self.imdb_job.is_finished, self.imdb_job
        return await self.tracker.get_ptp_metadata(self.imdb_id, key='title')

    def normalize_title(self, text):
        return text.strip()

    def validate_title(self, text):
        if not text:
            raise ValueError('Title must not be empty.')

    @functools.cached_property
    def year_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('year'),
            label='Year',
            precondition=self.make_precondition(
                'year_job',
                # Don't run this job if PTP already knows this movie.
                precondition=self.ptp_group_does_not_exist,
            ),
            prejobs=self.get_job_and_dependencies(
                # We need to wait for the PTP group ID to become available
                # before we can check if it exists.
                self.ptp_group_id_job,
            ),
            text=self.fetch_year,
            normalizer=self.normalize_year,
            validator=self.validate_year,
            **self.common_job_args(),
        )

    async def fetch_year(self):
        assert self.imdb_job.is_finished, self.imdb_job
        return await self.tracker.get_ptp_metadata(self.imdb_id, key='year')

    def normalize_year(self, text):
        return text.strip()

    def validate_year(self, text):
        if not text:
            raise ValueError('Year must not be empty.')
        try:
            year = int(text)
        except ValueError:
            raise ValueError('Year is not a number.')
        else:
            if not 1800 < year < datetime.now().year + 10:
                raise ValueError('Year is not reasonable.')

    @functools.cached_property
    def tags_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('tags'),
            label='Tags',
            precondition=self.make_precondition(
                'tags_job',
                # Don't run this job if PTP already knows this movie.
                precondition=self.ptp_group_does_not_exist,
            ),
            prejobs=self.get_job_and_dependencies(
                # We need to wait for the PTP group ID to become available
                # before we can check if it exists.
                self.ptp_group_id_job,
            ),
            text=self.fetch_tags,
            normalizer=self.normalize_tags,
            validator=self.validate_tags,
            **self.common_job_args(),
        )

    async def fetch_tags(self):
        assert self.imdb_job.is_finished, self.imdb_job
        return await self.tracker.get_ptp_metadata(self.imdb_id, key='tags')

    def normalize_tags(self, text):
        tags = [tag.strip() for tag in text.split(',')]
        deduped = sorted(dict.fromkeys(tags))
        return ', '.join(tag for tag in deduped if tag)

    def validate_tags(self, text):
        for tag in text.split(','):
            tag = tag.strip()
            if tag and tag not in metadata.ptp_tags:
                raise ValueError(f'Tag is not valid: {tag}')

    @functools.cached_property
    def cover_art_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('cover-art'),
            label='Cover Art',
            precondition=self.make_precondition(
                'cover_art_job',
                # Don't run this job if PTP already knows this movie.
                precondition=self.ptp_group_does_not_exist,
            ),
            prejobs=self.get_job_and_dependencies(
                # We need to wait for the PTP group ID to become available
                # before we can check if it exists.
                self.ptp_group_id_job,
            ),
            text=self.get_cover_art,
            validator=self.validate_cover_art,
            finish_on_success=True,
            **self.common_job_args(),
        )

    _cover_art_url_regex = re.compile(r'^https?://.+', flags=re.IGNORECASE)

    def validate_cover_art(self, text):
        if not self._cover_art_url_regex.search(text):
            raise ValueError('Must be an http(s):// URL.')

    async def get_cover_art(self):
        if self.options.get('cover_art', ''):
            return await self.get_cover_art_from_user()
        else:
            return await self.get_cover_art_from_ptp()

    async def get_cover_art_from_user(self):
        cover_art = self.options.get('cover_art', '')
        if self._cover_art_url_regex.search(cover_art):
            # User provided URL
            _log.debug('User provided cover art URL: %r', cover_art)
            return cover_art
        elif cover_art:
            # User provided file path
            _log.debug('User provided cover art file: %r', cover_art)
            # TODO: Use new ImageHostJob to make use of its fallback mechanism
            #       if one of image_hosts is down.
            try:
                return await self.image_hosts[0].upload(cover_art)
            except errors.RequestError as e:
                self.cover_art_job.error(e)
        return ''

    async def get_cover_art_from_ptp(self):
        # Try to get cover art URL from PTP API
        await self.imdb_job.wait()
        _log.debug('Getting cover art from PTP API: %r', self.imdb_id)
        return await self.tracker.get_ptp_metadata(self.imdb_id, key='art')

    @functools.cached_property
    def plot_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('plot'),
            label='Plot',
            precondition=self.make_precondition(
                'plot_job',
                # Don't run this job if PTP already knows this movie.
                precondition=self.ptp_group_does_not_exist,
            ),
            prejobs=self.get_job_and_dependencies(
                # We need to wait for the PTP group ID to become available
                # before we can check if it exists.
                self.ptp_group_id_job,
            ),
            text=self.fetch_plot,
            normalizer=self.normalize_plot,
            validator=self.validate_plot,
            finish_on_success=True,
            **self.common_job_args(),
        )

    async def fetch_plot(self):
        assert self.imdb_job.is_finished, self.imdb_job
        return await self.tracker.get_ptp_metadata(self.imdb_id, key='plot')

    def normalize_plot(self, text):
        return text.strip()

    def validate_plot(self, text):
        if not text:
            raise ValueError('Plot must not be empty.')

    @functools.cached_property
    def source_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('source'),
            label='Source',
            precondition=self.make_precondition('source_job'),
            text=self.autodetect_source,
            normalizer=self.normalize_source,
            validator=self.validate_source,
            finish_on_success=True,
            **self.common_job_args(),
        )

    _known_sources = {
        'Blu-ray': re.compile(r'^(?i:blu-?ray)$'),
        'DVD': re.compile(r'^(?i:dvd)'),  # DVD(Rip|5|9|...)
        'WEB': re.compile(r'^(?i:web)'),  # WEB(-DL|Rip)
        'HD-DVD': re.compile(r'^(?i:hd-?dvd)$'),
        'HDTV': re.compile(r'^(?i:hd-?tv)$'),
        'TV': re.compile(r'^(?i:tv)'),  # TV(Rip|...)
        'VHS': re.compile(r'^(?i:vhs)'),  # VHS(Rip|...)
    }

    def autodetect_source(self):
        # Find autodetected source in valid PTP sources
        source = self.release_name.source
        for known_source, regex in self._known_sources.items():
            if regex.match(source):
                # Finish the job without prompting the user
                return known_source

        # Let the user fix the autodetected source
        self.source_job.text = source

    def normalize_source(self, text):
        text = text.strip()
        for source in self._known_sources:
            if source.casefold() == text.casefold():
                return source
        return text

    def validate_source(self, text):
        if not text or text == 'UNKNOWN_SOURCE':
            raise ValueError('You must provide a source.')
        elif text not in self._known_sources:
            raise ValueError(f'Source is not valid: {text}')

    # # POST parameter "codec" seems to be optional.

    # @functools.cached_property
    # def video_codec_job(self):
    #     return jobs.dialog.TextFieldJob(
    #         name=self.get_job_name('video_codec'),
    #         label='Video Codec',
    #         precondition=self.make_precondition('video_codec_job'),
    #         text=self.autodetect_video_codec,
    #         finish_on_success=True,
    #         normalizer=self.normalize_video_codec,
    #         validator=self.validate_video_codec,
    #         **self.common_job_args(),
    #     )

    # def autodetect_video_codec(self):
    #     autodetected_video_format = self.release_name.video_format
    #     autodetected_source = self.release_name.source

    #     if autodetected_video_format == 'x264':
    #         return 'x264'
    #     elif autodetected_video_format == 'x265':
    #         return 'x265'

    #     elif autodetected_video_format == 'H.264':
    #         return 'H.264'
    #     elif autodetected_video_format == 'H.265':
    #         return 'H.265'

    #     elif autodetected_video_format == 'XviD':
    #         return 'XviD'
    #     elif autodetected_video_format == 'DivX':
    #         return 'DivX'

    #     elif autodetected_source == 'BD25':
    #         return 'BD25'
    #     elif autodetected_source == 'BD50':
    #         return 'BD50'
    #     elif autodetected_source == 'BD66':
    #         return 'BD66'
    #     elif autodetected_source == 'BD100':
    #         return 'BD100'

    #     elif autodetected_source == 'DVD5':
    #         return 'DVD5'
    #     elif autodetected_source == 'DVD9':
    #         return 'DVD9'

    #     else:
    #         self.video_codec_job.text = autodetected_video_format

    # def normalize_video_codec(self, text):
    #     return text.strip()

    # def validate_video_codec(self, text):
    #     if not text:
    #         raise ValueError('You must provide a video codec.')

    # # POST parameter "container" seems to be optional.

    # @functools.cached_property
    # def container_job(self):
    #     return jobs.dialog.TextFieldJob(
    #         name=self.get_job_name('container_codec'),
    #         label='Container',
    #         precondition=self.make_precondition('container_job'),
    #         text=self.autodetect_container,
    #         finish_on_success=True,
    #         normalizer=self.normalize_container,
    #         validator=self.validate_container,
    #         **self.common_job_args(),
    #     )

    # async def autodetect_container(self):
    #     if utils.fs.find_name('BDMV', self.content_path, validator=os.path.isdir):
    #         return 'm2ts'
    #     elif utils.fs.file_extension(self.content_path).lower() == 'iso':
    #         return 'ISO'

    #     info = utils.video.default_track('general', self.content_path, default=None)
    #     if info:
    #         container = info.get('Format')
    #         if container == 'Matroska':
    #             return 'MKV'
    #         elif container == 'MPEG-4':
    #             return 'MP4'
    #         elif container == 'MPEG-TS':
    #             return 'm2ts'
    #         elif container == 'MPEG-PS':
    #             return 'VOB IFO'
    #         elif container == 'AVI':
    #             return 'AVI'
    #         elif container == 'MPEG Video':
    #             return 'MPG'
    #         else:
    #             self.container_job.text = container

    # def normalize_container(self, text):
    #     return text.strip()

    # def validate_container(self, text):
    #     if not text:
    #         raise ValueError('You must provide a container format.')

    # # POST parameter "container" seems to be optional.

    # @functools.cached_property
    # def resolution_job(self):
    #     return jobs.dialog.TextFieldJob(
    #         name=self.get_job_name('resolution_codec'),
    #         label='Resolution',
    #         precondition=self.make_precondition('resolution_job'),
    #         text=self.autodetect_resolution,
    #         finish_on_success=True,
    #         normalizer=self.normalize_resolution,
    #         validator=self.validate_resolution,
    #         **self.common_job_args(),
    #     )

    # _known_resolutions = (
    #     '2160p',
    #     '1080p',
    #     '1080i',
    #     '720p',
    #     '576p',
    #     '480p',
    #     'PAL',
    #     'NTSC',
    # )

    # def autodetect_resolution(self):
    #     dvd_encoding = self.release_name.dvd_encoding
    #     if dvd_encoding:
    #         return dvd_encoding

    #     resolution = utils.video.resolution(self.content_path, default='')
    #     if resolution in self._known_resolutions:
    #         return resolution

    #     # Let the user fix the autodetected resolution
    #     self.resolution_job.text = resolution

    # def normalize_resolution(self, text):
    #     # WIDTHxHEIGHT
    #     match = re.search(r'^\s*(\d*)\s*[xX]\s*(\d*)\s*$', text)
    #     if match:
    #         width = match.group(1)
    #         height = match.group(2)
    #         return f'{width}x{height}'

    #     # 1080p, etc
    #     return text.strip()

    # def validate_resolution(self, text):
    #     if not text:
    #         raise ValueError('You must provide a video resolution.')
    #     elif (
    #             text not in self._known_resolutions
    #             and not re.search(r'^\d+x\d+$', text)
    #     ):
    #         raise ValueError('Invalid resolution')

    @property
    def post_data(self):
        post_data = self._post_data_common

        _log.debug('PTP group ID: %r', self.ptp_group_id)
        if self.ptp_group_id:
            _log.debug('Adding movie format to existing group')
            post_data.update(self._post_data_add_movie_format)
        else:
            _log.debug('Creating new movie group')
            post_data.update(self._post_data_add_new_movie)

        return post_data

    @property
    def _post_data_common(self):
        return {
            **{
                # Feature Film, Miniseries, Short Film, etc
                'type': self.get_job_attribute(self.type_job, 'choice'),

                # Mediainfo and Screenshots
                'release_desc': self.get_job_output(self.description_job, slice=0),

                # Is not main movie (bool)
                'special': '1' if self.options['not_main_movie'] else None,

                # Is personal rip (bool)
                'internalrip': '1' if self.options['personal_rip'] else None,

                # Is scene Release (bool)
                'scene': '1' if self.get_job_attribute(self.scene_check_job, 'is_scene_release') else None,

                # .nfo file content
                'nfo_text': self.read_nfo(),

                # Upload token from staff
                'uploadtoken': self.options.get('upload_token', None),
            },
            **self._post_data_common_source,
            **self._post_data_common_codec,
            **self._post_data_common_container,
            **self._post_data_common_resolution,
            **self._post_data_common_remaster,
            **self._post_data_common_subtitles,
            **self._post_data_common_trumpable,
        }

    @property
    def _post_data_common_source(self):
        return {
            'source': 'Other',
            'other_source': self.get_job_output(self.source_job, slice=0),
        }

    @property
    def _post_data_common_codec(self):
        return {
            # Server-side autodetection
            'codec': '* Auto-detect',
            'other_codec': '',

            # # Custom codec
            # 'codec': 'Other',
            # 'other_codec': self.get_job_output(self.video_codec_job, slice=0),
        }

    @property
    def _post_data_common_container(self):
        return {
            # Server-side autodetection
            'container': '* Auto-detect',
            'other_container': ''

            # # Custom container
            # 'container': 'Other',
            # 'other_container': self.get_job_output(self.container_job, slice=0),
        }

    @property
    def _post_data_common_resolution(self):
        return {
            # Server-side autodetection
            'resolution': '* Auto-detect',
            'other_resolution': '',
        }

        # Custom resolution
        # resolution = self.get_job_output(self.resolution_job, slice=0)
        # # WIDTHxHEIGHT
        # if 'x' in resolution:
        #     width, height = resolution.split('x', maxsplit=1)
        #     return {
        #         'resolution': 'Other',  # Custom resolution
        #         'other_resolution_width': width,
        #         'other_resolution_height': height,
        #     }

        # # 1080p, 720p, etc
        # else:
        #     return {
        #         'resolution': resolution,
        #         'other_resolution': releaseInfo.Resolution,
        #     }

    @property
    def _post_data_common_subtitles(self):
        # Translate BCP47 codes into PTP numeric codes from upload.php
        codes = []
        for sub in self.subtitles:
            ptp_code = None

            if ptp_code is None and sub.territory and sub.forced:
                # Language, country and "forced" flag, e.g. "pt-BR (forced)"
                ptp_code = metadata.ptp_subtitles.get(f'{sub.language}-{sub.territory} (forced)')

            if ptp_code is None and sub.territory:
                # Language and "forced" flag, e.g. "pt-BR"
                ptp_code = metadata.ptp_subtitles.get(f'{sub.language}-{sub.territory}')

            if ptp_code is None and sub.forced:
                # Language and "forced" flag, e.g. "pt (forced)"
                ptp_code = metadata.ptp_subtitles.get(f'{sub.language} (forced)')

            if ptp_code is None:
                # Language, e.g. "pt"
                ptp_code = metadata.ptp_subtitles.get(sub.language)

            if ptp_code:
                codes.append(ptp_code)

        # Explicitly send "No subtitles" if there aren't any
        if not codes and not self.subtitles:
            codes.append(metadata.ptp_subtitles['No Subtitles'])

        _log.debug('PTP subtitle codes: %r', codes)
        return {'subtitles[]': codes}

    @property
    def _post_data_common_remaster(self):
        return {
            # 'remaster_title': ...,  # Edition Information ("Director's Cut", "Dual Audio", etc.)
            # 'remaster_year': ...,
            # 'remaster_other_input': ...,
        }

    @property
    def _post_data_common_trumpable(self):
        return {
            'trumpable[]': [
                # Convert pretty strings to API values
                metadata.PtpTrumpableReason.from_string(string).value
                for string in self.trumpable_job.output
            ],
        }

    @property
    def _post_data_add_movie_format(self):
        # Upload another release to existing movie group
        return {
            'groupid': self.ptp_group_id,
        }

    @property
    def _post_data_add_new_movie(self):
        # Upload movie that is not on PTP yet in any format
        post_data = {
            # IMDb ID (must be 7 characters without the leading "tt")
            'imdb': self.tracker.normalize_imdb_id(self.imdb_id),
            # Release year
            'title': self.get_job_output(self.title_job, slice=0),
            # Release year
            'year': self.get_job_output(self.year_job, slice=0),
            # Movie plot or summary
            'album_desc': self.get_job_output(self.plot_job, slice=0),
            # Genre
            'tags': self.get_job_output(self.tags_job, slice=0),
            # Youtube ID
            # 'trailer': ...,
            # Poster URL
            'image': self.get_job_output(self.cover_art_job, slice=0),
        }

        # TODO: Is this required?
        # # ???: For some reason PtpUploader uses "artist" and "importance" in the
        # #      POST data while the website form uses "artists", "importances"
        # #      and "roles".
        # artists = self.get_job_output(self.artists_job):
        # if artists:
        #     post_data['artist'] = []
        #     post_data['importance'] = []
        #     for name in artists:
        #         # `importance` is the artist type:
        #         # 1: Director
        #         # 2: Writer
        #         # 3: Producer
        #         # 4: Composer
        #         # 5: Actor
        #         # 6: Cinematographer
        #         post_data['importance'].append('1')
        #         post_data['artist'].append(name)
        #         post_data['role'].append(name)

        return post_data



"""
CLI subcommands

A command provides a :attr:`~.CommandBase.jobs` property that returns a sequence
of :class:`~.jobs.base.JobBase` instances.

It is important that the jobs are only instantiated once and the
:attr:`~.CommandBase.jobs` property doesn't create new jobs every time it is
accessed. The easiest way to achieve this is with the
:func:`functools.cached_property` decorator.

CLI arguments are available as :attr:`~.CommandBase.args` and configuration
files as :attr:`~.CommandBase.config`. These should be used to create arguments
for jobs.

The docstrings of :class:`~.base.CommandBase` subclasses are used as the
description in the ``--help`` output.
"""

import functools

from ....utils import subclasses, submodules
from .base import CommandBase


@functools.lru_cache(maxsize=None)
def _register_commands():
    subcmdclss = sorted(subclasses(CommandBase, submodules(__package__)),
                        key=lambda subcmdcls: subcmdcls.names[0])
    for subcmdcls in subcmdclss:
        subcmdcls.register()


def run(args):
    """
    Instantiate :class:`~.base.CommandBase`

    This is just a thin wrapper around :meth:`~.base.CommandBase.run` that
    ensures all commands are registered first.

    :param args: CLI arguments
    """
    _register_commands()
    return CommandBase.run(args)



import os

import pytest


@pytest.fixture(scope='module')
def strict_filename_sanitization(module_mocker):
    # Allow this project to be cloned on Windows.
    def sanitize_filename(filename):
        for c in r'<>:"/\|?*':
            filename = filename.replace(c, '_')
        return filename

    module_mocker.patch('upsies.utils.fs.sanitize_filename', sanitize_filename)


@pytest.fixture(scope='session')
def data_dir():
    segments = __file__.split(os.sep)
    tests_index = segments[::-1].index('tests')
    tests_dir = os.sep.join(segments[:-tests_index])
    data_dir = os.path.join(tests_dir, 'data')
    if not os.path.exists(data_dir):
        os.mkdir(data_dir)
    return data_dir


# Don't make HTTP requests unless they are explicitly allowed.
# See tests/tools_test/dbs_test/conftest.py
def pytest_addoption(parser):
    parser.addoption('--allow-requests', action='store_true',
                     help='Use this option to update the request cache')


# By default, pytest-asyncio uses a new loop for every test, but utils.http uses
# module-level request asyncio.Lock() objects to avoid making the same request
# multiple times simultaneously. This fixture uses the same event_loop for all
# test modules in this package and closes it after all tests ran.
@pytest.fixture(scope='module', autouse=True)
def event_loop():
    import asyncio

    loop = asyncio.new_event_loop()
    yield loop
    loop.close()



import re
from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies import errors
from upsies.utils.predbs import predbclub


@pytest.fixture
def api():
    return predbclub.PredbclubApi()


def test_name():
    assert predbclub.PredbclubApi.name == 'predbclub'


def test_label():
    assert predbclub.PredbclubApi.label == 'PreDB.club'


def test_default_config():
    assert predbclub.PredbclubApi.default_config == {}


@pytest.mark.asyncio
async def test_search(api, mocker):
    mocker.patch.object(api, '_get_q', return_value='foo bar @team baz')
    mocker.patch.object(api, '_request_all_pages', AsyncMock(return_value=('Foo', 'Bar', 'Baz')))
    query = Mock(keywords=('kw1', 'kw2'), group='ASDF')
    results = await api._search(query)
    assert results == ('Foo', 'Bar', 'Baz')
    assert api._get_q.call_args_list == [call(('kw1', 'kw2'), 'ASDF')]
    assert api._request_all_pages.call_args_list == [call('foo bar @team baz')]


@pytest.mark.parametrize(
    argnames='keywords, group, exp_q',
    argvalues=(
        ((), None, ''),
        (('',), None, ''),
        ((' ', '', '  '), None, ''),
        (('foo', 'bar'), None, 'foo bar'),
        (('', '  '' foo ', '', '  ' 'b a r', '  ', '', ' ', 'baz', '', '  '), None, 'foo b a r baz'),
        (('foo', 'bar'), 'ASDF', 'foo bar @team asdf'),
        (('foo', 'bar'), 'M@G', r'foo bar @team m\@g'),
    ),
    ids=lambda v: str(v),
)
def test_get_q(keywords, group, exp_q, api):
    q = api._get_q(keywords, group)
    assert q == exp_q


@pytest.mark.parametrize('last_result_index', (5, 6, 7, 8, 9, 10))
@pytest.mark.asyncio
async def test_request_all_pages_stops_when_there_is_no_next_page(last_result_index, api, mocker):
    results_count = 100
    results_per_page = 3
    results = [f'Result {i}' for i in range(1, (results_count) + 1)]

    pages = []
    for i in range(0, len(results), results_per_page):
        page = results[i:i + results_per_page]
        if i >= last_result_index - 2:
            page_len = last_result_index % results_per_page
            page = page[:page_len]
            next_page = -1
            pages.append((page, next_page))
            break
        else:
            page = results[i:i + results_per_page]
            next_page = int((i / results_per_page) + 2)
            pages.append((page, next_page))

    mocker.patch.object(api, '_request_page', AsyncMock(side_effect=pages))

    exp_results = results[:last_result_index]

    mock_q = 'foo bar @team baz'
    return_value = await api._request_all_pages(mock_q)
    assert return_value == exp_results

    max_page_index_requested = (last_result_index // results_per_page) + 1
    exp_request_page_calls = [
        call(mock_q, i) for i in range(1, max_page_index_requested + 1)
    ]
    assert api._request_page.call_args_list == exp_request_page_calls

@pytest.mark.asyncio
async def test_request_all_pages_does_not_request_pages_indefinitely(api, mocker):
    results_count = 100
    results_per_page = 3
    results = [f'Result {i}' for i in range(1, (results_count) + 1)]

    pages = [
        (results[i:i + results_per_page], int((i / results_per_page) + 2))
        for i in range(0, len(results), results_per_page)
    ]
    mocker.patch.object(api, '_request_page', AsyncMock(side_effect=pages))

    max_page_index_requested = 30
    exp_results = results[:results_per_page * max_page_index_requested]

    mock_q = 'foo bar @team baz'
    return_value = await api._request_all_pages(mock_q)
    assert return_value == exp_results

    assert api._request_page.call_args_list == [
        call(mock_q, i) for i in range(1, max_page_index_requested + 1)
    ]


@pytest.mark.parametrize(
    argnames='q, page, response, exp_return_value, exp_exception',
    argvalues=(
        (
            'foo @team bar',
            1,
            {'status': 'success', 'data': {'rows': [{'name': f'Foo {i}'} for i in range(30)], 'reqCount': 29}},
            (tuple(f'Foo {i}' for i in range(30)), 2),
            None,
        ),
        (
            'foo @team bar',
            1,
            {'status': 'success', 'data': {'rows': [{'name': f'Foo {i}'} for i in range(30)], 'reqCount': 30}},
            (tuple(f'Foo {i}' for i in range(30)), 2),
            None,
        ),
        (
            'foo @team bar',
            1,
            {'status': 'success', 'data': {'rows': [{'name': f'Foo {i}'} for i in range(30)], 'reqCount': 31}},
            (tuple(f'Foo {i}' for i in range(30)), -1),
            None,
        ),
        (
            'foo @team bar',
            1,
            {'status': 'not success', 'message': 'Something went wrong'},
            None,
            errors.RequestError(f'{predbclub.PredbclubApi.label}: Something went wrong'),
        ),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_request_page(q, page, response, exp_return_value, exp_exception, api, mocker):
    mock_search_url = 'http://foo/api'
    mocker.patch.object(api, '_search_url', mock_search_url)

    get_mock = mocker.patch('upsies.utils.http.get', AsyncMock(return_value=Mock(
        json=Mock(return_value=response),
    )))

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await api._request_page(q, page)
    else:
        return_value = await api._request_page(q, page)
        assert return_value == exp_return_value

    exp_params = {'q': q, 'count': 100, 'page': page}
    assert get_mock.call_args_list == [call(mock_search_url, params=exp_params, cache=True)]


@pytest.mark.asyncio
async def test_release_files(api):
    with pytest.raises(NotImplementedError):
        await api._release_files('foo')



from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import errors
from upsies.jobs.imghost import ImageHostJob
from upsies.utils.imghosts import ImageHostBase


@pytest.fixture
def fooimg(tmp_path):
    class FooImageHost(ImageHostBase):
        name = 'foo.img'
        default_config = {}
        upload = AsyncMock()
        _upload_image = 'not an abstract base method'
    return FooImageHost(cache_directory=tmp_path / 'fooimg')

@pytest.fixture
def barimg(tmp_path):
    class BarImageHost(ImageHostBase):
        name = 'bar.img'
        default_config = {}
        upload = AsyncMock()
        _upload_image = 'not an abstract base method'
    return BarImageHost(cache_directory=tmp_path / 'barimg')

@pytest.fixture
def bazimg(tmp_path):
    class BazImageHost(ImageHostBase):
        name = 'baz.img'
        default_config = {}
        upload = AsyncMock()
        _upload_image = 'not an abstract base method'
    return BazImageHost(cache_directory=tmp_path / 'bazimg')

@pytest.fixture
def arfimg(tmp_path):
    class ArfImageHost(ImageHostBase):
        name = 'arf.img'
        default_config = {}
        upload = AsyncMock()
        _upload_image = 'not an abstract base method'
    return ArfImageHost(cache_directory=tmp_path / 'arfimg')

@pytest.fixture
def imghosts(fooimg, barimg, bazimg, arfimg):
    return fooimg, barimg, bazimg, arfimg

@pytest.fixture
def imghosts_map(fooimg, barimg, bazimg, arfimg):
    return {'fooimg': fooimg, 'barimg': barimg, 'bazimg': bazimg, 'arfimg': arfimg}

@pytest.fixture
async def make_ImageHostJob(tmp_path, imghosts):
    def make_ImageHostJob(home_directory=tmp_path, imghosts=imghosts, images_total=0, enqueue=()):
        return ImageHostJob(
            home_directory=home_directory,
            cache_directory=tmp_path,
            ignore_cache=False,
            imghosts=imghosts,
            images_total=images_total,
            enqueue=enqueue,
        )
    return make_ImageHostJob


def test_cache_id(make_ImageHostJob):
    job = make_ImageHostJob()
    assert job.cache_id is None


def test_initialize_sets_cache_directory_of_imghosts(make_ImageHostJob, fooimg, barimg):
    job = make_ImageHostJob(imghosts=(fooimg, barimg))
    for imghost in job._imghosts:
        assert imghost.cache_directory is job.cache_directory


def test_initialize_is_called_with_enqueue_and_images_total(make_ImageHostJob):
    with pytest.raises(RuntimeError, match=(r'^You must not give both arguments '
                                            r'"enqueue" and "images_total"\.$')):
        make_ImageHostJob(images_total=24, enqueue=('a', 'b', 'c'))


def test_initialize_is_called_without_enqueue_and_without_images_total(make_ImageHostJob, mocker):
    job = make_ImageHostJob()
    assert job.images_total == 0


async def test_initialize_is_called_with_enqueue(make_ImageHostJob, mocker):
    job = make_ImageHostJob(enqueue=('a', 'b', 'c'))
    assert job.images_total == 3


async def test_initialize_is_called_with_images_total(make_ImageHostJob, mocker):
    job = make_ImageHostJob(images_total=5)
    assert job.images_total == 5


@pytest.mark.parametrize(
    argnames='imghost_names, exp_mock_calls',
    argvalues=(
        ((), ['_upload_to_one']),
        (('fooimg',), ['_upload_to_one']),
        (('fooimg', 'barimg'), ['_upload_to_any']),
        (('fooimg', 'barimg', 'bazimg'), ['_upload_to_any']),
    ),
    ids=lambda v: repr(v),
)
async def test_handle_input(imghost_names, exp_mock_calls, imghosts_map, make_ImageHostJob, mocker):
    image_path = 'path/to/some.jpg'
    imghosts = [imghosts_map[name] for name in imghost_names]
    job = make_ImageHostJob(imghosts=imghosts)
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, '_upload_to_one'), '_upload_to_one')
    mocks.attach_mock(mocker.patch.object(job, '_upload_to_any'), '_upload_to_any')

    exp_mock_calls = [
        getattr(call, method_name)(image_path)
        for method_name in exp_mock_calls
    ]
    return_value = await job.handle_input(image_path)
    assert return_value is None
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='number_of_imghosts, upload_exception, exp_upload_call, exp_error_call',
    argvalues=(
        (0, None, False, False),
        (1, None, True, False),
        (0, errors.RequestError('foo'), False, False),
        (1, errors.RequestError('foo'), True, True),
    ),
    ids=lambda v: repr(v),
)
async def test__upload_to_one(
        number_of_imghosts, upload_exception, exp_upload_call, exp_error_call,
        imghosts_map, make_ImageHostJob, mocker,
):
    image_path = 'path/to/some.jpg'
    imghosts = list(imghosts_map.values())[:number_of_imghosts]
    job = make_ImageHostJob(imghosts=imghosts)
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, '_upload', side_effect=upload_exception), '_upload')
    mocks.attach_mock(mocker.patch.object(job, 'error'), 'error')

    exp_mock_calls = []
    if exp_upload_call:
        exp_mock_calls.append(call._upload(image_path, imghosts[0]))
    if exp_error_call:
        exp_mock_calls.append(call.error(f'{imghosts[0].name}: Upload failed: some.jpg: {upload_exception}'))

    return_value = await job._upload_to_one(image_path)
    assert return_value is None
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='number_of_imghosts, upload_exceptions, exp_upload_calls, exp_warn_calls, exp_error_call',
    argvalues=(
        (4, [], 1, 0, False),
        (4, [
            errors.RequestError('This service is down.'),
        ], 2, 1, False),
        (4, [
            errors.RequestError('This service is down.'),
            errors.RequestError('This service is also down.'),
        ], 3, 2, False),
        (4, [
            errors.RequestError('This service is down.'),
            errors.RequestError('This service is also down.'),
            errors.RequestError('Another service is down.'),
        ], 4, 3, False),
        (4, [
            errors.RequestError('This service is down.'),
            errors.RequestError('This service is also down.'),
            errors.RequestError('Another service is down.'),
            errors.RequestError('Last service is down.'),
        ], 4, 4, True),
    ),
    ids=lambda v: repr(v),
)
async def test__upload_to_any(
        number_of_imghosts, upload_exceptions, exp_upload_calls, exp_warn_calls, exp_error_call,
        imghosts_map, make_ImageHostJob, mocker,
):
    image_path = 'path/to/some.jpg'
    imghosts = list(imghosts_map.values())[:number_of_imghosts]
    job = make_ImageHostJob(imghosts=imghosts)
    mocks = Mock()
    upload_side_effects = upload_exceptions + [None for _ in range(exp_upload_calls)]
    mocks.attach_mock(mocker.patch.object(job, '_upload', side_effect=upload_side_effects), '_upload')
    mocks.attach_mock(mocker.patch.object(job, 'warn'), 'warn')
    mocks.attach_mock(mocker.patch.object(job, 'error'), 'error')

    exp_mock_calls = []
    if exp_upload_calls:
        for i, imghost in enumerate(imghosts[:exp_upload_calls]):
            exp_mock_calls.append(call._upload(image_path, imghost))
            is_final_host = i == len(imghosts[:exp_upload_calls])
            if not is_final_host:
                try:
                    e = upload_exceptions[i]
                except IndexError:
                    pass
                else:
                    exp_mock_calls.append(call.warn(f'{imghost.name}: Upload failed: some.jpg: {e}'))

    if exp_error_call:
        exp_mock_calls.append(call.error('All upload attempts failed.'))

    for c in exp_mock_calls:
        print('exp_call:', c)

    return_value = await job._upload_to_any(image_path)
    assert return_value is None
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='ignore_cache, exp_cache_argument',
    argvalues=(
        (True, False),
        (False, True),
    ),
    ids=lambda v: str(v),
)
async def test__upload(ignore_cache, exp_cache_argument, fooimg, make_ImageHostJob, mocker):
    job = make_ImageHostJob(imghosts=(fooimg,))
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(fooimg, 'upload', side_effect=lambda path, cache: f'http://{path}'), 'upload')
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'send')
    mocker.patch.object(type(job), 'ignore_cache', PropertyMock(return_value=ignore_cache))

    image_paths = [
        f'path/to/{i}.jpg'
        for i in range(6)
    ]
    for image_path in image_paths:
        return_value = await job._upload(image_path, fooimg)
        assert return_value is None
        assert mocks.mock_calls[-2] == call.upload(image_path, cache=exp_cache_argument)
        assert mocks.mock_calls[-1] == call.send(f'http://{image_path}')
        assert job._uploaded_images[-1] == f'http://{image_path}'
    assert len(job._uploaded_images) == len(image_paths)


@pytest.mark.parametrize(
    argnames='is_finished, images_uploaded, images_total, exp_exit_code',
    argvalues=(
        (False, -1, -1, None),
        (True, 0, 0, 1),
        (True, 1, 2, 1),
        (True, 2, 2, 0),
    ),
    ids=lambda v: str(v),
)
def test_exit_code(is_finished, images_uploaded, images_total, exp_exit_code, make_ImageHostJob, mocker):
    job = make_ImageHostJob()
    mocker.patch.object(type(job), 'is_finished', PropertyMock(return_value=is_finished))
    mocker.patch.object(type(job), 'images_uploaded', PropertyMock(return_value=images_uploaded))
    mocker.patch.object(type(job), 'images_total', PropertyMock(return_value=images_total))
    assert job.exit_code == exp_exit_code


def test_uploaded_images(make_ImageHostJob, mocker):
    job = make_ImageHostJob()
    mocker.patch.object(job, '_uploaded_images', ['<uploaded 1.jpg>', '<uploaded 2.jpg>', '<uploaded 3.jpg>'])
    assert job.uploaded_images == ('<uploaded 1.jpg>', '<uploaded 2.jpg>', '<uploaded 3.jpg>')


def test_urls_by_file(make_ImageHostJob, mocker):
    job = make_ImageHostJob()
    mocker.patch.object(job, '_urls_by_file', {
        '1.jpg': 'http://1.jpg',
        '2.jpg': 'http://2.jpg',
        '3.jpg': 'http://3.jpg',
    })
    urls_by_file = job.urls_by_file
    assert urls_by_file == {
        '1.jpg': 'http://1.jpg',
        '2.jpg': 'http://2.jpg',
        '3.jpg': 'http://3.jpg',
    }
    urls_by_file.clear()
    assert urls_by_file == {}
    assert job.urls_by_file == {
        '1.jpg': 'http://1.jpg',
        '2.jpg': 'http://2.jpg',
        '3.jpg': 'http://3.jpg',
    }


def test_images_uploaded(make_ImageHostJob, mocker):
    job = make_ImageHostJob()
    mocker.patch.object(job, '_uploaded_images', ['<uploaded 1.jpg>', '<uploaded 2.jpg>', '<uploaded 3.jpg>'])
    assert job.images_uploaded == 3


def test_images_total(make_ImageHostJob, mocker):
    job = make_ImageHostJob()
    unique_object = object()
    mocker.patch.object(job, '_images_total', unique_object)
    assert job.images_total is unique_object

    job.images_total = 123.5
    assert job.images_total == 123
    with pytest.raises(ValueError):
        job.images_total = 'foo'

    job.set_images_total(124)
    assert job.images_total == 124
    with pytest.raises(TypeError):
        job.set_images_total(object())



import asyncio
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import errors
from upsies.jobs import webdb
from upsies.utils.webdbs import Query, WebDbApiBase


@pytest.fixture
def foodb(mocker):
    class FooDb(WebDbApiBase):
        name = 'foodb'
        label = 'FooDB'
        default_config = {}
        sanitize_query = Mock(return_value=Query('sanitized query'))
        get_id_from_text = Mock(return_value=None)
        search = AsyncMock()
        cast = AsyncMock()
        creators = AsyncMock()
        _countries = AsyncMock()
        directors = AsyncMock()
        genres = AsyncMock()
        poster_url = AsyncMock()
        rating = AsyncMock()
        rating_min = 0
        rating_max = 10
        _runtimes = AsyncMock()
        summary = AsyncMock()
        _title_original = AsyncMock()
        _titles_english = AsyncMock()
        type = AsyncMock()
        url = AsyncMock()
        year = AsyncMock()
    return FooDb()


@pytest.fixture
def job(foodb, tmp_path, mocker):
    mocker.patch('upsies.jobs.webdb._Searcher', Mock(return_value=Mock(wait=AsyncMock())))
    mocker.patch('upsies.jobs.webdb._InfoUpdater', Mock(return_value=Mock(wait=AsyncMock())))
    job = webdb.WebDbSearchJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        db=foodb,
        query=Query('Mock Title'),
    )
    assert job._db is foodb
    return job


def test_WebDbSearchJob_name(job):
    assert job.name == 'foodb-id'


def test_WebDbSearchJob_label(job):
    assert job.label == 'FooDB ID'


def test_WebDbSearchJob_cache_id(job):
    assert job.cache_id == job.query


def test_WebDbSearchJob_db(job):
    assert job.db is job._db


def test_WebDbSearchJob_query(job):
    assert job.query is job._query


@pytest.mark.parametrize(
    argnames='query, exp_from_any_called',
    argvalues=(
        (Query('a query instance'), False),
        ('not a query instance', True),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio  # Ensure aioloop exists
async def test_WebDbSearchJob_initialize_sets_query(query, exp_from_any_called, foodb, mocker, tmp_path):
    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.jobs.webdb.webdbs.Query.from_any'), 'from_any')
    mocks.attach_mock(foodb.sanitize_query, 'sanitize_query')
    mocks.sanitize_query.return_value = Query('sanitized query')

    job = webdb.WebDbSearchJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        db=foodb,
        query=query,
    )

    if exp_from_any_called:
        exp_mock_calls = [
            call.from_any(query),
            call.sanitize_query(mocks.from_any.return_value),
        ]
    else:
        exp_mock_calls = [
            call.sanitize_query(query),
        ]
    assert mocks.mock_calls == exp_mock_calls
    assert job._query is mocks.sanitize_query.return_value
    assert job._query.signal.signals['changed'] == [
        job._handle_query_changed,
    ]

@pytest.mark.parametrize(
    argnames='no_id_ok, exp_no_id_ok',
    argvalues=(
        (True, True),
        (False, False),
        (None, False),
        (1, True),
    ),
)
@pytest.mark.asyncio  # Ensure aioloop exists
async def test_WebDbSearchJob_initialize_sets_no_id_ok(no_id_ok, exp_no_id_ok, tmp_path, foodb):
    job = webdb.WebDbSearchJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        db=foodb,
        query=Query(title='The Foo', year=2010),
        no_id_ok=no_id_ok,
    )
    assert job.no_id_ok == exp_no_id_ok
    assert foodb.sanitize_query.call_args_list == [
        call(Query(title='The Foo', year=2010)),
    ]

@pytest.mark.asyncio  # Ensure aioloop exists
async def test_WebDbSearchJob_initialize_creates_searcher(tmp_path, mocker, foodb):
    Searcher_mock = mocker.patch('upsies.jobs.webdb._Searcher', Mock())
    job = webdb.WebDbSearchJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        db=foodb,
        query='path/to/foo',
    )
    assert job._searcher is Searcher_mock.return_value
    assert Searcher_mock.call_args_list == [call(
        search_coro=foodb.search,
        results_callback=job._handle_search_results,
        error_callback=job.warn,
        exception_callback=job.exception,
        searching_callback=job._handle_searching_status,
    )]

@pytest.mark.asyncio  # Ensure aioloop exists
async def test_WebDbSearchJob_initialize_creates_info_updater(tmp_path, mocker, foodb):
    InfoUpdater_mock = mocker.patch('upsies.jobs.webdb._InfoUpdater', Mock())
    make_update_info_func_mock = mocker.patch(
        'upsies.jobs.webdb.WebDbSearchJob._make_update_info_func',
        Mock(
            side_effect=(
                'id func',
                'summary func',
                'title_original func',
                'title_english func',
                'genres func',
                'directors func',
                'cast func',
                'countries func',
            ),
        )
    )
    job = webdb.WebDbSearchJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        db=foodb,
        query='path/to/foo',
    )
    assert job._info_updater is InfoUpdater_mock.return_value
    assert InfoUpdater_mock.call_args_list == [call(
        error_callback=job.warn,
        exception_callback=job.exception,
        targets={
            'id': 'id func',
            'summary': 'summary func',
            'title_original': 'title_original func',
            'title_english': 'title_english func',
            'genres': 'genres func',
            'directors': 'directors func',
            'cast': 'cast func',
            'countries': 'countries func',
        },
    )]
    assert make_update_info_func_mock.call_args_list == [
        call('id'),
        call('summary'),
        call('title_original'),
        call('title_english'),
        call('genres'),
        call('directors'),
        call('cast'),
        call('countries'),
    ]


@pytest.mark.asyncio  # Ensure aioloop exists
async def test_WebDbSearchJob_make_update_info_func(tmp_path, mocker, foodb):
    mocker.patch('upsies.jobs.webdb.WebDbSearchJob._update_info')
    job = webdb.WebDbSearchJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        db=foodb,
        query='path/to/foo',
    )
    func = job._make_update_info_func('key')
    func('value 1')
    assert job._update_info.call_args_list == [call('key', 'value 1')]
    func('value 2')
    assert job._update_info.call_args_list == [call('key', 'value 1'), call('key', 'value 2')]


def test_WebDbSearchJob_execute_does_initial_search(job, mocker):
    mocker.patch.object(job, 'search')
    assert job.search.call_args_list == []
    job.execute()
    assert job.search.call_args_list == [call(job.query)]


@pytest.mark.asyncio
async def test_WebDbSearchJob_wait(job):
    assert job._searcher.wait.call_args_list == []
    assert job._info_updater.wait.call_args_list == []
    asyncio.get_running_loop().call_soon(job.finish)
    await job.wait()
    assert job.is_finished
    assert job._searcher.wait.call_args_list == [call()]
    assert job._info_updater.wait.call_args_list == [call()]


def test_WebDbSearchJob_finish(job):
    job._searcher = Mock()
    job._info_updater = Mock()
    job.finish()
    assert job._searcher.cancel.call_args_list == [call()]
    assert job._info_updater.cancel.call_args_list == [call()]


def test_WebDbSearchJob_search_before_finished(job, mocker):
    job._query = Mock(name='original query')

    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.jobs.webdb.webdbs.Query.from_any'), 'from_any')
    mocks.attach_mock(mocker.patch.object(job._query, 'update'), 'update')
    mocks.attach_mock(mocker.patch.object(job._db, 'sanitize_query'), 'sanitize_query')
    mocks.attach_mock(mocker.patch.object(job._searcher, 'search'), 'search')
    mocks.attach_mock(mocker.patch.object(job, 'clear_warnings'), 'clear_warnings')
    job.signal.register('query_updated', mocks.query_updated)

    query_id = id(job.query)
    job.search('new query')
    assert id(job.query) == query_id

    assert mocks.mock_calls == [
        call.clear_warnings(),
        call.from_any('new query'),
        call.sanitize_query(mocks.from_any.return_value),
        call.update(mocks.sanitize_query.return_value, silent=True),
        call.search(job._query),
        call.query_updated(job._query),
    ]


def test_WebDbSearchJob_search_after_finished(job, mocker):
    job._query = Mock(name='original query')

    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.jobs.webdb.webdbs.Query.from_any'), 'from_any')
    mocks.attach_mock(mocker.patch.object(job._query, 'update'), 'update')
    mocks.attach_mock(mocker.patch.object(job._db, 'sanitize_query'), 'sanitize_query')
    mocks.attach_mock(mocker.patch.object(job._searcher, 'search'), 'search')
    mocks.attach_mock(mocker.patch.object(job, 'clear_warnings'), 'clear_warnings')
    job.signal.register('query_updated', mocks.query_updated)

    query_id = id(job.query)
    job.finish()
    job.search('new query')
    assert id(job.query) == query_id

    assert mocks.mock_calls == []


def test_WebDbSearchJob_handle_query_changed(job, mocker):
    mocker.patch.object(job, 'search')
    job._handle_query_changed('mock query')
    assert job.search.call_args_list == [call('mock query')]


def test_WebDbSearchJob_searching_status(job):
    cb = Mock()
    job.signal.register('searching_status', cb)
    assert job.is_searching is False
    job._handle_searching_status(True)
    assert job.is_searching is True
    assert cb.call_args_list == [call(True)]
    job._handle_searching_status(False)
    assert job.is_searching is False
    assert cb.call_args_list == [call(True), call(False)]


def test_WebDbSearchJob_search_results(job):
    results = ('foo', 'bar', 'baz')
    cb = Mock()
    job.signal.register('search_results', cb)
    job._handle_search_results(results)
    assert job._info_updater.set_result.call_args_list == [call(results[0])]
    assert cb.call_args_list == [call(results)]

def test_WebDbSearchJob_no_search_results(job):
    results = ()
    cb = Mock()
    job.signal.register('search_results', cb)
    job._handle_search_results(results)
    assert job._info_updater.set_result.call_args_list == [call(None)]
    assert cb.call_args_list == [call(results)]


def test_WebDbSearchJob_update_info_emits_info_updating_signal(job):
    cb = Mock()
    job.signal.register('info_updating', cb.info_updating)
    job.signal.register('info_updated', cb.info_updated)
    job._update_info('The Key', Ellipsis)
    assert cb.info_updating.call_args_list == [call('The Key')]
    assert cb.info_updated.call_args_list == []

def test_WebDbSearchJob_update_info_emits_info_updated_signal(job):
    cb = Mock()
    job.signal.register('info_updating', cb.info_updating)
    job.signal.register('info_updated', cb.info_updated)
    job._update_info('The Key', 'The Info')
    assert cb.info_updating.call_args_list == []
    assert cb.info_updated.call_args_list == [call('The Key', 'The Info')]


def test_WebDbSearchJob_result_focused(job):
    job.result_focused('The Result')
    assert job._info_updater.set_result.call_args_list == [call('The Result')]


@pytest.mark.parametrize(
    argnames='result, is_searching, no_id_ok, exp_is_finished, exp_output',
    argvalues=(
        (Mock(id='mock id'), False, False, True, ('mock id',)),
        (Mock(id='mock id'), False, True, True, ('mock id',)),
        (Mock(id='mock id'), True, False, False, ()),
        (Mock(id='mock id'), True, True, False, ()),
        (None, False, False, False, ()),
        (None, False, True, True, ()),
        (None, True, False, False, ()),
        (None, True, True, False, ()),
    ),
)
def test_WebDbSearchJob_result_selected(result, is_searching, no_id_ok, exp_is_finished, exp_output, job):
    assert not job.is_finished
    job._is_searching = is_searching
    job.no_id_ok = no_id_ok
    job.result_selected(result)
    assert job.is_finished is exp_is_finished
    assert job.output == exp_output


@pytest.mark.parametrize(
    argnames='value, exp_value',
    argvalues=(
        (True, True),
        (False, False),
        (None, False),
        (1, True),
    ),
)
@pytest.mark.asyncio  # Ensure aioloop exists
async def test_WebDbSearchJob_no_id_ok(value, exp_value, job):
    assert job.no_id_ok is False
    job.no_id_ok = value
    assert job.no_id_ok is exp_value


@pytest.mark.parametrize(
    argnames='parent_exit_code, no_id_ok, exp_exit_code',
    argvalues=(
        (None, False, None),
        (None, True, None),
        (0, False, 0),
        (0, True, 0),
        (1, False, 1),
        (1, True, 0),
    ),
)
@pytest.mark.asyncio  # Ensure aioloop exists
async def test_WebDbSearchJob_exit_code(parent_exit_code, no_id_ok, exp_exit_code, mocker, job):
    mocker.patch('upsies.jobs.base.JobBase.exit_code', PropertyMock(return_value=parent_exit_code))
    job.no_id_ok = no_id_ok
    assert job.exit_code is exp_exit_code


@pytest.fixture
async def searcher():
    searcher = webdb._Searcher(
        search_coro=AsyncMock(),
        results_callback=Mock(),
        error_callback=Mock(),
        exception_callback=Mock(),
        searching_callback=Mock(),
    )
    yield searcher


@pytest.mark.asyncio
async def test_Searcher_wait_while_not_searching(searcher):
    searcher._search_task = None
    await searcher.wait()
    assert searcher._search_task is None

@pytest.mark.asyncio
async def test_Searcher_wait_while_searching(searcher):
    searcher._search_task = asyncio.create_task(AsyncMock()())
    await searcher.wait()
    assert searcher._search_task.done()


def test_Searcher_cancel_while_searching(searcher):
    cancel_mock = Mock()
    searcher._search_task = Mock(cancel=cancel_mock)
    searcher.cancel()
    assert cancel_mock.call_args_list == [call()]

def test_Searcher_cancel_while_not_searching(searcher):
    searcher._search_task = None
    searcher.cancel()
    assert searcher._search_task is None


@pytest.mark.asyncio
async def test_Searcher_search_while_not_searching(searcher, mocker):
    mocker.patch.object(searcher, '_search', return_value=AsyncMock)
    searcher._search_task = None
    mock_query = Mock()
    searcher.search(mock_query)
    assert searcher._search.call_args_list == [call(mock_query)]
    assert isinstance(searcher._search_task, asyncio.Task)
    assert not searcher._search_task.done()
    assert not searcher._search_task.cancelled()
    assert searcher._current_query == mock_query.copy.return_value
    assert mock_query.copy.call_args_list == [call()]

@pytest.mark.asyncio
async def test_Searcher_search_while_searching(searcher, mocker):
    mocker.patch.object(searcher, '_search', return_value=AsyncMock)
    old_task = searcher._search_task = Mock()
    mock_query = Mock()
    searcher.search(mock_query)
    assert searcher._search.call_args_list == [call(mock_query)]
    assert isinstance(searcher._search_task, asyncio.Task)
    new_task = searcher._search_task
    assert new_task is not old_task
    assert old_task.cancel.call_args_list == [call()]
    assert not new_task.done()
    assert not new_task.cancelled()

@pytest.mark.asyncio
async def test_Searcher_search_does_not_search_for_current_query(searcher, mocker):
    mocker.patch.object(searcher, '_search', AsyncMock())
    searcher._current_query = mock_query = Mock()
    searcher._search_task = mock_search_task = Mock()
    searcher.search(mock_query)
    await asyncio.sleep(0.1)
    assert searcher._search_task is mock_search_task
    assert searcher._search_task.mock_calls == []
    assert searcher._search.call_args_list == []
    assert searcher._current_query.copy.call_args_list == []

@pytest.mark.asyncio
async def test_Searcher_search_reports_exception(searcher, mocker):
    exception_cb = Mock()
    mocker.patch.object(searcher, '_exception_callback', exception_cb)
    mocker.patch.object(searcher, '_search', AsyncMock(side_effect=TypeError('your code sucks')))
    mock_query = Mock()
    searcher.search(mock_query)
    await asyncio.sleep(0.1)
    assert isinstance(exception_cb.call_args_list[0][0][0], TypeError)
    assert str(exception_cb.call_args_list[0][0][0]) == 'your code sucks'

@pytest.mark.asyncio
async def test_Searcher_search_ignores_CancelledError(searcher, mocker):
    exception_cb = Mock()
    mocker.patch.object(searcher, '_exception_callback', exception_cb)
    mocker.patch.object(searcher, '_search', lambda query: asyncio.sleep(10))
    mock_query = Mock()
    searcher.search(mock_query)
    searcher.cancel()
    await asyncio.sleep(0.1)
    assert exception_cb.call_args_list == []


@pytest.mark.asyncio
async def test_Searcher__search_reports_results(searcher, mocker):
    mock_methods = {
        '_results_callback': Mock(),
        '_searching_callback': Mock(),
        # '_delay': AsyncMock(),
        '_search_coro': AsyncMock(return_value=('mock result 1', 'mock result 2')),
        '_error_callback': Mock(),
    }
    mocks = Mock(**mock_methods)
    mocker.patch.multiple(searcher, **mock_methods)
    await searcher._search('mock query')
    assert mocks.mock_calls == [
        call._results_callback(()),
        call._searching_callback(True),
        # call._delay(),
        call._search_coro('mock query'),
        call._searching_callback(False),
        call._results_callback(('mock result 1', 'mock result 2')),
    ]

@pytest.mark.asyncio
async def test_Searcher__search_reports_error(searcher, mocker):
    mock_methods = {
        '_results_callback': Mock(),
        '_searching_callback': Mock(),
        # '_delay': AsyncMock(),
        '_search_coro': AsyncMock(side_effect=errors.RequestError('internet is down')),
        '_error_callback': Mock(),
    }
    mocks = Mock(**mock_methods)
    mocker.patch.multiple(searcher, **mock_methods)
    await searcher._search('mock query')
    assert mocks.mock_calls == [
        call._results_callback(()),
        call._searching_callback(True),
        # call._delay(),
        call._search_coro('mock query'),
        call._error_callback(errors.RequestError('internet is down')),
        call._searching_callback(False),
        call._results_callback(()),
    ]

# TODO: _delay() is disabled for now. If this doesn't cause any issues, _delay()
#       below should be removed completely.

# @pytest.mark.asyncio
# async def test_Searcher_delay_sleeps(searcher, mocker):
#     mocker.patch('upsies.jobs.webdb.time_monotonic', Mock(side_effect=(1003, 1003.001)))
#     sleep_mock = mocker.patch('asyncio.sleep', new_callable=AsyncMock)
#     searcher._min_seconds_between_searches = 5
#     searcher._previous_search_time = 1000
#     await searcher._delay()
#     assert sleep_mock.call_args_list == [call(2)]
#     assert searcher._previous_search_time == 1003.001

# @pytest.mark.asyncio
# async def test_Searcher_delay_does_not_sleep(searcher, mocker):
#     mocker.patch('upsies.jobs.webdb.time_monotonic', Mock(side_effect=(1006, 1006.001)))
#     sleep_mock = mocker.patch('asyncio.sleep', new_callable=AsyncMock)
#     searcher._min_seconds_between_searches = 5
#     searcher._previous_search_time = 1000
#     await searcher._delay()
#     assert sleep_mock.call_args_list == []
#     assert searcher._previous_search_time == 1006.001


@pytest.fixture
async def info_updater():
    info_updater = webdb._InfoUpdater(
        targets={},
        error_callback=Mock(),
        exception_callback=Mock(),
    )
    yield info_updater


@pytest.mark.asyncio
async def test_InfoUpdater_cancel_while_not_updating(info_updater):
    info_updater._update_task = None
    info_updater.cancel()
    await asyncio.sleep(0)
    assert info_updater._update_task is None

@pytest.mark.asyncio
async def test_InfoUpdater_cancel_while_updating(info_updater):
    info_updater._update_task = asyncio.ensure_future(asyncio.sleep(10))
    info_updater.cancel()
    await info_updater.wait()
    assert info_updater._update_task.cancelled()


@pytest.mark.asyncio
async def test_InfoUpdater_wait_while_not_updating(info_updater):
    info_updater._update_task = None
    await info_updater.wait()
    assert info_updater._update_task is None

@pytest.mark.asyncio
async def test_InfoUpdater_wait_while_updating(info_updater):
    info_updater._update_task = asyncio.ensure_future(asyncio.sleep(0.1))
    await info_updater.wait()
    assert info_updater._update_task.done()


@pytest.mark.asyncio
async def test_InfoUpdater_set_result_reports_exception(info_updater, mocker):
    info_updater._update = AsyncMock(side_effect=TypeError('your code sucks'))
    info_updater.set_result('mock result')
    await asyncio.sleep(0.1)
    assert isinstance(info_updater._exception_callback.call_args_list[0][0][0], TypeError)
    assert str(info_updater._exception_callback.call_args_list[0][0][0]) == 'your code sucks'

@pytest.mark.asyncio
async def test_InfoUpdater_set_result_cancels_previous_update_task(info_updater, mocker):
    old_update_task = info_updater._update_task = asyncio.ensure_future(asyncio.sleep(10))
    info_updater.set_result('mock result')
    await asyncio.sleep(0)  # _update_task.cancel() to propagate
    assert old_update_task.cancelled()
    assert info_updater._update_task is not old_update_task

@pytest.mark.asyncio
async def test_InfoUpdater_set_result_does_not_cancel_previous_update_task(info_updater, mocker):
    info_updater._update_task = None
    info_updater.set_result('mock result')
    assert isinstance(info_updater._update_task, asyncio.Task)

@pytest.mark.asyncio
async def test_InfoUpdater_set_result_sets_result(info_updater, mocker):
    info_updater.set_result('mock result 1')
    assert info_updater._result == 'mock result 1'
    task1 = info_updater._update_task
    assert isinstance(task1, asyncio.Task)
    info_updater.set_result('mock result 2')
    assert info_updater._result == 'mock result 2'
    task2 = info_updater._update_task
    assert isinstance(task2, asyncio.Task)
    assert task1 is not task2

@pytest.mark.asyncio
async def test_InfoUpdater_set_result_calls_targets_with_no_result_set(info_updater, mocker):
    cbs = Mock()
    info_updater._targets = {
        'title': cbs.title,
        'year': cbs.year,
        'genre': cbs.genre,
    }
    info_updater.set_result(None)
    assert cbs.mock_calls == [
        call.title(''),
        call.year(''),
        call.genre(''),
    ]
    assert info_updater._update_task is None

@pytest.mark.asyncio
async def test_InfoUpdater_set_result_calls_targets_with_result_set(info_updater, mocker):
    mock_update = AsyncMock()
    mocker.patch.object(info_updater, '_update', mock_update)
    cbs = Mock()
    info_updater._targets = {
        'title': cbs.title,
        'year': cbs.year,
        'genre': cbs.genre,
    }
    result = Mock(
        title='The Foo',
        year='2015',
        genre=AsyncMock(return_value='bar'),
    )
    old_update_task = info_updater._update_task
    info_updater.set_result(result)
    assert cbs.mock_calls == [
        call.title('The Foo'),
        call.year('2015'),
    ]
    assert old_update_task is not info_updater._update_task
    assert isinstance(info_updater._update_task, asyncio.Task)
    assert mock_update.call_args_list == [call()]


@pytest.mark.asyncio
async def test_InfoUpdater_update(info_updater, mocker):
    mocker.patch.object(info_updater, '_call_callback', Mock())
    info_updater._targets = {
        'title': Mock(),
        'year': Mock(),
        'genre': Mock(),
        'summary': Mock(),
    }
    info_updater._result = Mock(
        id='123',
        title='The Foo',
        year='2015',
        genre=AsyncMock(return_value=('She', 'He', 'Doggo')),
        summary=AsyncMock(return_value='Something happens.'),
    )
    info_updater._call_callback.side_effect = (
        info_updater._result.genre(),
        info_updater._result.summary(),
    )
    await info_updater._update()
    assert info_updater._call_callback.call_args_list == [
        call(
            cache_key=('123', 'genre'),
            value_getter=info_updater._result.genre,
            callback=info_updater._targets['genre'],
        ),
        call(
            cache_key=('123', 'summary'),
            value_getter=info_updater._result.summary,
            callback=info_updater._targets['summary'],
        ),
    ]
    assert info_updater._targets['title'].call_args_list == []
    assert info_updater._targets['year'].call_args_list == []
    assert info_updater._targets['genre'].call_args_list == []
    assert info_updater._targets['summary'].call_args_list == []


@pytest.mark.asyncio
async def test_UpdateInfoThread_call_callback_gets_value_from_value_getter(info_updater, mocker):
    mocks = Mock(
        value_getter=AsyncMock(return_value='The Value'),
        callback=Mock(),
        sleep_mock=AsyncMock(),
    )
    mocker.patch('asyncio.sleep', mocks.sleep_mock)
    info_updater._cache.clear()
    await info_updater._call_callback(
        cache_key=('id', 'key'),
        value_getter=mocks.value_getter,
        callback=mocks.callback,
    )
    assert mocks.mock_calls == [
        call.callback(Ellipsis),
        call.sleep_mock(info_updater._delay_between_updates),
        call.value_getter(),
        call.callback('The Value'),
    ]
    assert info_updater._cache[('id', 'key')] == 'The Value'

@pytest.mark.asyncio
async def test_UpdateInfoThread_call_callback_gets_value_from_cache(info_updater, mocker):
    mocks = Mock(
        value_getter=AsyncMock(return_value='The Value'),
        callback=Mock(),
        sleep_mock=AsyncMock(),
    )
    mocker.patch('asyncio.sleep', mocks.sleep_mock)
    info_updater._cache[('id', 'key')] = 'The Cached Value'
    await info_updater._call_callback(
        callback=mocks.callback,
        value_getter=mocks.value_getter,
        cache_key=('id', 'key'),
    )
    assert mocks.mock_calls == [
        call.callback('The Cached Value'),
    ]
    assert info_updater._cache[('id', 'key')] == 'The Cached Value'

@pytest.mark.asyncio
async def test_UpdateInfoThread_call_callback_handles_RequestError(info_updater, mocker):
    mocks = Mock(
        value_getter=AsyncMock(side_effect=errors.RequestError('Nah')),
        callback=Mock(),
        error_callback=Mock(),
        sleep_mock=AsyncMock(),
    )
    info_updater._error_callback = mocks.error_callback
    mocker.patch('asyncio.sleep', mocks.sleep_mock)
    info_updater._cache.clear()
    await info_updater._call_callback(
        callback=mocks.callback,
        value_getter=mocks.value_getter,
        cache_key=('id', 'key'),
    )
    assert mocks.mock_calls == [
        call.callback(Ellipsis),
        call.sleep_mock(info_updater._delay_between_updates),
        call.value_getter(),
        call.callback(''),
        call.error_callback(errors.RequestError('Nah')),
    ]
    assert info_updater._cache == {}


def test_value_as_string_with_string(info_updater):
    assert info_updater._value_as_string('foo bar baz') == 'foo bar baz'

def test_value_as_string_with_iterable(info_updater):
    assert info_updater._value_as_string(('foo', 'bar', 'baz')) == 'foo, bar, baz'
    assert info_updater._value_as_string(['foo', 2, {3}]) == 'foo, 2, {3}'

def test_value_as_string_with_other_type(info_updater):
    assert info_updater._value_as_string(type) == str(type)



"""
Generate properly formatted release name
"""

import functools

from .... import errors, jobs, utils
from .base import CommandBase


class release_name(CommandBase):
    """
    Generate properly formatted release name

    IMDb is searched to get the correct title, year and alternative title if
    applicable.

    Audio and video information is detected with mediainfo.

    Missing required information is highlighted with placeholders,
    e.g. "UNKNOWN_RESOLUTION".
    """

    names = ('release-name', 'rn')

    argument_definitions = {
        'RELEASE': {
            'type': utils.argtypes.release,
            'help': 'Release name or path to release content',
        },
        ('--separator', '-s'): {
            'default': ' ',
            'help': 'Separator between parts (default: " ")',
        },
    }

    @functools.cached_property
    def jobs(self):
        return (self.imdb_job, self.release_name_job)

    @functools.cached_property
    def release_name(self):
        return utils.release.ReleaseName(
            self.args.RELEASE,
            separator=self.args.separator,
        )

    @functools.cached_property
    def release_name_job(self):
        return jobs.dialog.TextFieldJob(
            name='release-name',
            label='Release Name',
            text=str(self.release_name),
            home_directory=self.home_directory,
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
        )

    def _update_release_name(self, imdb_id):
        self.release_name_job.attach_task(
            self.release_name_job.fetch_text(
                coro=self.release_name.fetch_info(webdb=self.imdb, webdb_id=imdb_id),
                nonfatal_exceptions=(
                    errors.RequestError,
                ),
                default=str(self.release_name),
            ),
        )

    @functools.cached_property
    def imdb_job(self):
        # To be able to fetch the original title, year, etc, we need to ask for
        # an ID first. IMDb seems to be best.
        return jobs.webdb.WebDbSearchJob(
            home_directory=self.home_directory,
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
            query=utils.webdbs.Query.from_path(self.args.RELEASE),
            db=self.imdb,
            callbacks={'output': self._update_release_name},
        )

    @functools.cached_property
    def imdb(self):
        return utils.webdbs.webdb('imdb')



from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies.utils.predbs import corruptnet


@pytest.fixture
def api():
    return corruptnet.CorruptnetApi()


def test_name():
    assert corruptnet.CorruptnetApi.name == 'corruptnet'


def test_label():
    assert corruptnet.CorruptnetApi.label == 'pre.corrupt-net.org'


def test_default_config():
    assert corruptnet.CorruptnetApi.default_config == {}


@pytest.mark.parametrize(
    argnames='keywords, group, exp_return_value',
    argvalues=(
        (['foo ', ' bar', 'b az'], None, 'foo bar b az'),
        (['foo', 'bar', 'baz'], ' NOGROUP  ', 'foo bar baz group:NOGROUP'),
        (['foo', 'bar', 'baz', '  s03E06 '], 'NOGROUP', 'foo bar baz s03E06 group:NOGROUP'),
    ),
)
def test_join_keywords(keywords, group, exp_return_value, api):
    return_value = api._join_keywords(keywords, group)
    assert return_value == exp_return_value


@pytest.mark.parametrize(
    argnames='query, results_per_page, max_pages, pages, exp_return_value, exp_mock_calls',
    argvalues=(
        pytest.param(
            Mock(keywords=['my', 'keywords'], group='MYGROUP'),
            3, 4,
            [['a', 'b', 'c'], ['d', 'e', 'f'], ['g', 'h', 'i'], ['j', 'k', 'l'], ['m', 'n', 'o'], ['p', 'q', 'r'], []],
            ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l'],
            [
                call._request_page(['my', 'keywords'], 'MYGROUP', 0),
                call._request_page(['my', 'keywords'], 'MYGROUP', 3),
                call._request_page(['my', 'keywords'], 'MYGROUP', 6),
                call._request_page(['my', 'keywords'], 'MYGROUP', 9),
            ],
            id='Maximum number of pages',
        ),

        pytest.param(
            Mock(keywords=['my', 'keywords'], group='MYGROUP'),
            3, 4,
            [['a', 'b', 'c'], ['d', 'e', 'f'], []],
            ['a', 'b', 'c', 'd', 'e', 'f'],
            [
                call._request_page(['my', 'keywords'], 'MYGROUP', 0),
                call._request_page(['my', 'keywords'], 'MYGROUP', 3),
                call._request_page(['my', 'keywords'], 'MYGROUP', 6),
            ],
            id='Results count divisible by page count',
        ),

        pytest.param(
            Mock(keywords=['my', 'keywords'], group='MYGROUP'),
            3, 4,
            [['a', 'b', 'c'], ['d', 'e', 'f'], ['g']],
            ['a', 'b', 'c', 'd', 'e', 'f', 'g'],
            [
                call._request_page(['my', 'keywords'], 'MYGROUP', 0),
                call._request_page(['my', 'keywords'], 'MYGROUP', 3),
                call._request_page(['my', 'keywords'], 'MYGROUP', 6),
            ],
            id='Results count not divisible by page count',
        ),
    ),
)
@pytest.mark.asyncio
async def test_search(query, results_per_page, max_pages, pages, exp_return_value, exp_mock_calls, api, mocker):
    mocks = Mock()
    mocker.patch.object(api, '_max_pages', max_pages)
    mocker.patch.object(api, '_results_per_page', results_per_page)
    mocks.attach_mock(
        mocker.patch.object(api, '_request_page', side_effect=pages),
        '_request_page',
    )

    return_value = await api._search(query)
    assert return_value == exp_return_value
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='keywords, group, page, exp_params',
    argvalues=(
        pytest.param(
            ['my', 'keywords'], 'MYGROUP', 0,
            {'search': 'my keywords group:MYGROUP', 'ts': 0, 'pretimezone': 0, 'timezone': 0},
            id='First page',
        ),
        pytest.param(
            ['my', 'keywords'], 'MYGROUP', 123,
            {'search': 'my keywords group:MYGROUP', 'ts': 0, 'pretimezone': 0, 'timezone': 0, 'page': 123},
            id='Not first page',
        ),
    ),
)
@pytest.mark.asyncio
async def test_request_page(keywords, group, page, exp_params, api, mocker):
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch('upsies.utils.predbs.corruptnet.http.get', AsyncMock()),
        'get',
    )
    mocks.attach_mock(
        mocker.patch.object(api, '_yield_release_names', Mock()),
        '_yield_release_names',
    )

    return_value = await api._request_page(keywords, group, page)
    assert return_value == mocks._yield_release_names.return_value
    assert mocks.mock_calls == [
        call.get(api._search_url, params=exp_params, cache=True, verify=False),
        call._yield_release_names(mocks.get.return_value),
    ]


@pytest.mark.parametrize(
    argnames='response, exp_return_value',
    argvalues=(
        pytest.param(
            (
                '<tr>'
                '<td>X264</td>'
                '<td>&nbsp;This.2003.1080p.BluRay.x264-<span><font>TEHGRP</font></span></td>'
                '<td>foo</td>'
                '</tr>'
                '<tr>'
                '<td>X264</td>'
                '<td>&nbsp;That.2004.720p.DVDRip.x264-<span><font>ASDF</font></span></td>'
                '<td>foo</td>'
                '</tr>'
                '<tr>'
                '<td>X264</td>'
                '<td>&nbsp;Some.Of.These.1999.720p.WEB.x264-<span><font>ARF</font></span></td>'
                '<td>foo</td>'
                '</tr>'
            ),
            [
                'This.2003.1080p.BluRay.x264-TEHGRP',
                'That.2004.720p.DVDRip.x264-ASDF',
                'Some.Of.These.1999.720p.WEB.x264-ARF',
            ],
            id='Results found',
        ),
        pytest.param(
            (
                '<td>X264</td>'
                '<td>&nbsp;This.2003.1080p.BluRay.x264-<span><font>TEHGRP</font></span></td>'
                '<td>foo</td>'
            ),
            [],
            id='No results found: No <tr> tag',
        ),
        pytest.param(
            (
                '<tr>'
                '<p>X264</p>'
                '<div>&nbsp;This.2003.1080p.BluRay.x264-<span><font>TEHGRP</font></span></div>'
                '</tr>'
            ),
            [],
            id='No results found: No <td> tag',
        ),
    ),
)
def test_yield_release_name(response, exp_return_value, api):
    return_value = list(api._yield_release_names(response))
    assert return_value == exp_return_value


@pytest.mark.asyncio
async def test_release_files(api):
    with pytest.raises(NotImplementedError):
        await api._release_files('foo')



import os
import re
from unittest.mock import call

import pytest

from upsies import errors
from upsies.utils import image, video

VF_FLAGS = (
    'full_chroma_int',
    'full_chroma_inp',
    'accurate_rnd',
    'spline',
)

@pytest.mark.parametrize(
    argnames='video_file, timestamp, screenshot_file, color_matrix, exp_args',
    argvalues=(
        pytest.param(
            'video.mkv', 123, 'foo%.png', 'BT.2020',
            (
                '-y', '-hide_banner', '-loglevel', 'error',
                '-ss', '123',
                '-i', 'ffmpeg_input:video.mkv',
                '-vf', ':'.join((
                    'scale=in_h_chr_pos=0',
                    'in_v_chr_pos=0',
                    'in_color_matrix=bt2020',
                    'flags=' + '+'.join(VF_FLAGS),
                )),
                '-pix_fmt', 'rgb24', '-vframes', '1',
                'file:foo%%.png',
            ),
            id='BT.2020',
        ),
        pytest.param(
            'video.mkv', 123, 'foo%.png', 'BT.709',
            (
                '-y', '-hide_banner', '-loglevel', 'error',
                '-ss', '123',
                '-i', 'ffmpeg_input:video.mkv',
                '-vf', ':'.join((
                    "scale='max(sar,1)*iw':'max(1/sar,1)*ih'",
                    'in_h_chr_pos=0',
                    'in_v_chr_pos=128',
                    'in_color_matrix=bt709',
                    'flags=' + '+'.join(VF_FLAGS),
                )),
                '-pix_fmt', 'rgb24', '-vframes', '1',
                'file:foo%%.png',
            ),
            id='BT.709',
        ),
        pytest.param(
            'video.mkv', 123, 'foo%.png', 'BT.601',
            (
                '-y', '-hide_banner', '-loglevel', 'error',
                '-ss', '123',
                '-i', 'ffmpeg_input:video.mkv',
                '-vf', ':'.join((
                    "scale='max(sar,1)*iw':'max(1/sar,1)*ih'",
                    'in_h_chr_pos=0',
                    'in_v_chr_pos=128',
                    'in_color_matrix=bt601',
                    'flags=' + '+'.join(VF_FLAGS),
                )),
                '-pix_fmt', 'rgb24', '-vframes', '1',
                'file:foo%%.png',
            ),
            id='BT.601',
        ),
        pytest.param(
            'video.mkv', 123, 'foo%.png', 'BT.foo',
            (
                '-y', '-hide_banner', '-loglevel', 'error',
                '-ss', '123',
                '-i', 'ffmpeg_input:video.mkv',
                '-vf', ':'.join((
                    "scale='max(sar,1)*iw':'max(1/sar,1)*ih'",
                    'flags=' + '+'.join(VF_FLAGS),
                )),
                '-pix_fmt', 'rgb24', '-vframes', '1',
                'file:foo%%.png',
            ),
            id='No color matrix',
        ),
    ),
    ids=lambda v: str(v),
)
def test_make_screenshot_cmd(video_file, timestamp, screenshot_file, color_matrix, exp_args, mocker):
    mocker.patch('upsies.utils.image._ffmpeg_executable', return_value='ffmpeg.executable')
    mocker.patch('upsies.utils.video.make_ffmpeg_input', side_effect=lambda path: f'ffmpeg_input:{path}')
    mocker.patch('upsies.utils.video.is_bt601', return_value=color_matrix == 'BT.601')
    mocker.patch('upsies.utils.video.is_bt709', return_value=color_matrix == 'BT.709')
    mocker.patch('upsies.utils.video.is_bt2020', return_value=color_matrix == 'BT.2020')

    cmd = image._make_screenshot_cmd(video_file, timestamp, screenshot_file)
    assert cmd == ('ffmpeg.executable',) + exp_args


def test_nonexisting_video_file(mocker):
    assert_file_readable_mock = mocker.patch('upsies.utils.fs.assert_file_readable',
                                             side_effect=errors.ContentError('Foo you'))
    sanitize_path_mock = mocker.patch('upsies.utils.fs.sanitize_path', return_value='sanitized path')
    run_mock = mocker.patch('upsies.utils.subproc.run')
    mock_file = 'path/to/foo.mkv'

    with pytest.raises(errors.ScreenshotError, match=r'^Foo you$'):
        image.screenshot(mock_file, 123, 'image.png')
    assert assert_file_readable_mock.call_args_list == [call(mock_file)]
    assert sanitize_path_mock.call_args_list == []
    assert run_mock.call_args_list == []


@pytest.mark.parametrize('invalid_timestamp', ('anywhere', [1, 2, 3]))
def test_invalid_timestamp(invalid_timestamp, mocker):
    mocker.patch('upsies.utils.fs.assert_file_readable', return_value=True)
    sanitize_path_mock = mocker.patch('upsies.utils.fs.sanitize_path', return_value='sanitized path')
    run_mock = mocker.patch('upsies.utils.subproc.run')

    mock_file = 'path/to/foo.mkv'
    with pytest.raises(errors.ScreenshotError, match=rf"^Invalid timestamp: {re.escape(repr(invalid_timestamp))}$"):
        image.screenshot(mock_file, invalid_timestamp, 'image.png')
    assert sanitize_path_mock.call_args_list == []
    assert run_mock.call_args_list == []

@pytest.mark.parametrize(
    argnames='timestamp',
    argvalues=(
        '12',
        '01:24',
        '1:02:03',
        '01:02:03',
        '123:02:03',
        123,
    ),
)
def test_valid_timestamp(timestamp, mocker):
    assert_file_readable_mock = mocker.patch('upsies.utils.fs.assert_file_readable', return_value=True)
    sanitize_path_mock = mocker.patch('upsies.utils.fs.sanitize_path', return_value='sanitized path')
    path_exists_mock = mocker.patch('os.path.exists', side_effect=(False, True))
    duration_mock = mocker.patch('upsies.utils.video.duration', return_value=1e6)
    make_screenshot_cmd_mock = mocker.patch('upsies.utils.image._make_screenshot_cmd', return_value='mock cmd')
    run_mock = mocker.patch('upsies.utils.subproc.run')

    mock_file = 'path/to/foo.mkv'
    screenshot_path = image.screenshot(mock_file, timestamp, 'image.png')
    assert screenshot_path == 'sanitized path'
    assert assert_file_readable_mock.call_args_list == [call(mock_file)]
    assert sanitize_path_mock.call_args_list == [call('image.png')]
    assert path_exists_mock.call_args_list == [call('sanitized path'), call('sanitized path')]
    assert duration_mock.call_args_list == [call(mock_file)]
    assert make_screenshot_cmd_mock.call_args_list == [call(mock_file, timestamp, 'sanitized path')]
    assert run_mock.call_args_list == [call(
        'mock cmd',
        ignore_errors=True,
        join_stderr=True,
    )]


def test_timestamp_after_video_end(mocker):
    assert_file_readable_mock = mocker.patch('upsies.utils.fs.assert_file_readable', return_value=True)
    sanitize_path_mock = mocker.patch('upsies.utils.fs.sanitize_path', return_value='sanitized path')
    path_exists_mock = mocker.patch('os.path.exists', side_effect=(False, True))
    duration_mock = mocker.patch('upsies.utils.video.duration', return_value=600)
    make_screenshot_cmd_mock = mocker.patch('upsies.utils.image._make_screenshot_cmd', return_value='mock cmd')
    run_mock = mocker.patch('upsies.utils.subproc.run')

    mock_file = 'path/to/foo.mkv'
    timestamp = 601
    with pytest.raises(errors.ScreenshotError, match=r'^Timestamp is after video end \(0:10:00\): 0:10:01$'):
        image.screenshot(mock_file, timestamp, 'image.png')
    assert assert_file_readable_mock.call_args_list == [call(mock_file)]
    assert sanitize_path_mock.call_args_list == [call('image.png')]
    assert path_exists_mock.call_args_list == [call('sanitized path')]
    assert duration_mock.call_args_list == [call(mock_file)]
    assert make_screenshot_cmd_mock.call_args_list == []
    assert run_mock.call_args_list == []


def test_getting_duration_fails(mocker):
    assert_file_readable_mock = mocker.patch('upsies.utils.fs.assert_file_readable', return_value=True)
    sanitize_path_mock = mocker.patch('upsies.utils.fs.sanitize_path', return_value='sanitized path')
    path_exists_mock = mocker.patch('os.path.exists', side_effect=(False, True))
    duration_mock = mocker.patch('upsies.utils.video.duration', side_effect=errors.ContentError('not a video file'))
    make_screenshot_cmd_mock = mocker.patch('upsies.utils.image._make_screenshot_cmd', return_value='mock cmd')
    run_mock = mocker.patch('upsies.utils.subproc.run')

    mock_file = 'path/to/foo.mkv'
    with pytest.raises(errors.ScreenshotError, match=r'^not a video file$'):
        image.screenshot(mock_file, 123, 'image.png')
    assert assert_file_readable_mock.call_args_list == [call(mock_file)]
    assert sanitize_path_mock.call_args_list == [call('image.png')]
    assert path_exists_mock.call_args_list == [call('sanitized path')]
    assert duration_mock.call_args_list == [call(mock_file)]
    assert make_screenshot_cmd_mock.call_args_list == []
    assert run_mock.call_args_list == []


def test_existing_screenshot_file(mocker):
    assert_file_readable_mock = mocker.patch('upsies.utils.fs.assert_file_readable', return_value=True)
    sanitize_path_mock = mocker.patch('upsies.utils.fs.sanitize_path', return_value='sanitized path')
    path_exists_mock = mocker.patch('os.path.exists', side_effect=(True, True))
    duration_mock = mocker.patch('upsies.utils.video.duration', return_value=1e6)
    make_screenshot_cmd_mock = mocker.patch('upsies.utils.image._make_screenshot_cmd', return_value='mock cmd')
    run_mock = mocker.patch('upsies.utils.subproc.run')

    mock_file = 'path/to/foo.mkv'
    screenshot_path = image.screenshot(mock_file, '1:02:03', 'image.png')
    assert screenshot_path == 'sanitized path'
    assert assert_file_readable_mock.call_args_list == [call(mock_file)]
    assert sanitize_path_mock.call_args_list == [call('image.png')]
    assert path_exists_mock.call_args_list == [call('sanitized path')]
    assert duration_mock.call_args_list == []
    assert make_screenshot_cmd_mock.call_args_list == []
    assert run_mock.call_args_list == []


def test_overwrite_existing_screenshot_file(mocker):
    assert_file_readable_mock = mocker.patch('upsies.utils.fs.assert_file_readable', return_value=True)
    sanitize_path_mock = mocker.patch('upsies.utils.fs.sanitize_path', return_value='sanitized path')
    path_exists_mock = mocker.patch('os.path.exists', side_effect=(True, True))
    duration_mock = mocker.patch('upsies.utils.video.duration', return_value=1e6)
    make_screenshot_cmd_mock = mocker.patch('upsies.utils.image._make_screenshot_cmd', return_value='mock cmd')
    run_mock = mocker.patch('upsies.utils.subproc.run')

    mock_file = 'path/to/foo.mkv'
    screenshot_path = image.screenshot(mock_file, '1:02:03', 'image.png', overwrite=True)
    assert screenshot_path == 'sanitized path'
    assert assert_file_readable_mock.call_args_list == [call(mock_file)]
    assert sanitize_path_mock.call_args_list == [call('image.png')]
    assert path_exists_mock.call_args_list == [call('sanitized path')]
    assert duration_mock.call_args_list == [call(mock_file)]
    assert make_screenshot_cmd_mock.call_args_list == [call(mock_file, '1:02:03', 'sanitized path')]
    assert run_mock.call_args_list == [call(
        'mock cmd',
        ignore_errors=True,
        join_stderr=True,
    )]


def test_screenshot_file_does_not_exist_for_some_reason(mocker):
    assert_file_readable_mock = mocker.patch('upsies.utils.fs.assert_file_readable', return_value=True)
    sanitize_path_mock = mocker.patch('upsies.utils.fs.sanitize_path', return_value='sanitized path')
    path_exists_mock = mocker.patch('os.path.exists', side_effect=(False, False))
    duration_mock = mocker.patch('upsies.utils.video.duration', return_value=1e6)
    make_screenshot_cmd_mock = mocker.patch('upsies.utils.image._make_screenshot_cmd', return_value='mock cmd')
    run_mock = mocker.patch('upsies.utils.subproc.run', return_value='ffmpeg output')

    mock_file = 'path/to/foo.mkv'
    with pytest.raises(errors.ScreenshotError, match=r'^path/to/foo.mkv: Failed to create screenshot at 601: ffmpeg output$'):
        image.screenshot(mock_file, 601, 'image.png')
    assert assert_file_readable_mock.call_args_list == [call(mock_file)]
    assert sanitize_path_mock.call_args_list == [call('image.png')]
    assert path_exists_mock.call_args_list == [call('sanitized path'), call('sanitized path')]
    assert duration_mock.call_args_list == [call(mock_file)]
    assert make_screenshot_cmd_mock.call_args_list == [call(mock_file, 601, 'sanitized path')]
    assert run_mock.call_args_list == [call(
        'mock cmd',
        ignore_errors=True,
        join_stderr=True,
    )]


def test_screenshot_has_display_aspect_ratio(data_dir, tmp_path):
    video_file = os.path.join(data_dir, 'video', 'aspect_ratio.mkv')
    screenshot_file = str(tmp_path / 'image.jpg')
    screenshot_path = image.screenshot(video_file, 0, screenshot_file)
    assert screenshot_path == screenshot_file
    tracks = video._tracks(screenshot_file)
    width = int(tracks['Image'][0]['Width'])
    height = int(tracks['Image'][0]['Height'])
    assert 1279 <= width <= 1280
    assert height == 534



"""
HTTP methods with caching
"""

import asyncio
import builtins
import collections
import http
import io
import json
import os
import pathlib
import random
import re
import ssl
import time

import httpx

from .. import __project_name__, __version__, constants, errors, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)

_default_timeout = 60
_default_headers = {
    'Accept-Language': 'en-US,en;q=0.5',
    'User-Agent': f'{__project_name__}/{__version__}',
}

# Use one Lock per requested url + parameters so we can make multiple identical
# requests concurrently without bugging the server.
_request_locks = collections.defaultdict(lambda: asyncio.Lock())

# Map domain names to dictionaries of session cookies
_session_cookies = collections.defaultdict(lambda: {})


cache_directory = None
"""
Where to store cached responses

If this is set to a falsy value, default to
:attr:`~.constants.DEFAULT_CACHE_DIRECTORY`.
"""

def _get_cache_directory():
    return cache_directory or constants.DEFAULT_CACHE_DIRECTORY


async def get(
        url,
        headers={},
        params={},
        auth=None,
        cache=False,
        max_cache_age=float('inf'),
        user_agent=False,
        follow_redirects=True,
        verify=True,
        timeout=_default_timeout,
        cookies=None,
        debug_file=None,
    ):
    """
    Perform HTTP GET request

    :param str url: URL to request
    :param dict headers: Custom headers (added to default headers)
    :param dict params: Key value pairs passed in the URL,
        e.g. ``{"k": "v"}`` → ``http://host/get?k=v``
    :param auth: Basic access authentication; sequence of <username> and
        <password> or `None`
    :param bool cache: Whether to use cached response if available
    :param int,float max_cache_age: Maximum age of cache in seconds
    :param bool user_agent: Whether to send the User-Agent header
    :param bool follow_redirects: Whether to follow redirects
    :param bool verify: Whether to verify TLS connection or ignore any errors
        like expired certificate
    :param int,float timeout: Maximum number of seconds the request may take
    :param cookies: Cookies to include in the request (merged with
        existing cookies in the global client session)

        * If `cookies` is a :class:`dict`, forget them after the request.

        * If `cookies` is a :class:`str`, load cookies from that path (if it
          exists), perform the request, and save the new set of cookies to the
          same path. If this is a relative path, `cache_directory` is prepended.

        .. note:: Session cookies are handled separately and automatically.

    :return: Response text
    :rtype: Response
    :raise RequestError: if the request fails for any expected reason
    """
    return await _request(
        method='GET',
        url=url,
        headers=headers,
        params=params,
        auth=auth,
        cache=cache,
        max_cache_age=max_cache_age,
        user_agent=user_agent,
        follow_redirects=follow_redirects,
        verify=verify,
        timeout=timeout,
        cookies=cookies,
        debug_file=debug_file,
    )

async def post(
        url,
        headers={},
        data={},
        files={},
        auth=None,
        cache=False,
        max_cache_age=float('inf'),
        user_agent=False,
        follow_redirects=True,
        verify=True,
        timeout=_default_timeout,
        cookies=None,
        debug_file=None,
    ):
    """
    Perform HTTP POST request

    :param str url: URL to request
    :param dict headers: Custom headers (added to default headers)
    :param dict data: Data to send as application/x-www-form-urlencoded
    :param dict files: Files to send as multipart/form-data as a dictionary that
        maps field names to file paths or dictionaries. For dictionaries, these
        keys are used:

             ``file``
                 File path or file-like object, e.g. return value of
                 :func:`open` or :class:`~.io.BytesIO`.

             ``filename`` (optional)
                 Name of the file that is reported to the server. Defaults to
                 :func:`~.os.path.basename` of ``file`` if ``file`` is a string,
                 `None` otherwise.

             ``mimetype`` (optional)
                 MIME type of the file content. If not provided and `file` is a
                 string, guess based on extension of ``file``. If `None`, do not
                 send any MIME type.

        Examples:

        >>> files = {
        >>>     "image": "path/to/foo.jpg",
        >>>     "document": {
        >>>         "file": "path/to/bar",
        >>>         "filename": "something.txt",
        >>>         "mimetype": "text/plain",
        >>>     },
        >>>     "data": {
        >>>         "file": io.BytesIO(b"binary data"),
        >>>         "mimetype": "application/octet-stream",
        >>>      },
        >>>     "more_data": io.BytesIO(b"more binary data"),
        >>> }
    :param auth: Basic access authentication; sequence of <username> and
        <password> or `None`
    :param bool cache: Whether to use cached response if available
    :param int,float max_cache_age: Maximum age of cache in seconds
    :param bool user_agent: Whether to send the User-Agent header
    :param bool follow_redirects: Whether to follow redirects
    :param bool verify: Whether to verify TLS connection or ignore any errors
        like expired certificate
    :param int,float timeout: Maximum number of seconds the request may take
    :param cookies: Cookies to include in the request (merged with existing
        cookies in the global client session); see :func:`get`

    :return: Response text
    :rtype: Response
    :raise RequestError: if the request fails for any expected reason
    """
    return await _request(
        method='POST',
        url=url,
        headers=headers,
        data=data,
        files=files,
        auth=auth,
        cache=cache,
        max_cache_age=max_cache_age,
        user_agent=user_agent,
        follow_redirects=follow_redirects,
        verify=verify,
        timeout=timeout,
        cookies=cookies,
        debug_file=debug_file,
    )

async def download(url, filepath, *args, cache=False, **kwargs):
    """
    Write downloaded data to file

    :param url: Where to download the data from
    :param filepath: Where to save the downloaded data

    Any other arguments are passed to :func:`get`, except for `cache`, which is
    always `False`.

    If `filepath` exists, no request is made.

    :raise RequestError: if anything goes wrong
    :return: `filepath`
    """
    if not os.path.exists(filepath) or not cache:
        _log.debug('Downloading %r to %r', url, filepath)
        response = await get(url, *args, cache=False, **kwargs)
        try:
            with open(filepath, 'wb') as f:
                f.write(response.bytes)
        except OSError as e:
            msg = e.strerror if e.strerror else str(e)
            raise errors.RequestError(f'Unable to write {filepath}: {msg}')
    else:
        _log.debug('Already downloaded %r to %r', url, filepath)
    return filepath


class Response(str):
    """
    Response to an HTTP request

    This is a subclass of :class:`str` with additional attributes and methods.
    """

    def __new__(cls, text, bytes, headers=None, status_code=None):
        obj = super().__new__(cls, text)
        obj._bytes = bytes
        obj._headers = headers if headers is not None else {}
        obj._status_code = status_code
        return obj

    @property
    def headers(self):
        """HTTP headers"""
        return self._headers

    @property
    def status_code(self):
        """HTTP status code"""
        return self._status_code

    @property
    def bytes(self):
        """Response data as :class:`bytes`"""
        return self._bytes

    def json(self):
        """
        Parse the response text as JSON

        :raise RequestError: if parsing fails
        """
        try:
            return json.loads(self)
        except ValueError as e:
            if self:
                raise errors.RequestError(f'Malformed JSON: {str(self)!r}: {e}')
            else:
                raise errors.RequestError('Malformed JSON: Empty string')

    def __repr__(self):
        kwargs = [
            f'text={str(self)!r}',
            f'bytes={self.bytes!r}',
        ]
        if self.headers != {}:
            kwargs.append(f'headers={self.headers!r}')
        if self.status_code is not None:
            kwargs.append(f'status_code={self.status_code!r}')
        return f'{type(self).__name__}({", ".join(kwargs)})'


async def _request(
        method,
        url,
        headers={},
        params={},
        data={},
        files={},
        auth=None,
        cache=False,
        max_cache_age=float('inf'),
        user_agent=False,
        follow_redirects=True,
        verify=True,
        timeout=_default_timeout,
        cookies=None,
        debug_file=None,
    ):
    if method.upper() not in ('GET', 'POST'):
        raise ValueError(f'Invalid method: {method}')

    client = httpx.AsyncClient(
        headers={**_default_headers, **headers},
        cookies=_load_permanent_cookies(cookies),
        verify=bool(verify),
    )
    async with client:
        # Create request object
        if isinstance(data, (bytes, str)):
            build_request_args = {'content': data}
        else:
            # Remove fields with a value of `None`
            build_request_args = {'data': {
                k: v
                for k, v in data.items()
                if v is not None
            }}
        request = client.build_request(
            method=str(method),
            url=str(url),
            cookies=_load_session_cookies(httpx.URL(url).host),
            params=params,
            files=_open_files(files),
            timeout=timeout,
            **build_request_args,
        )

        # Adjust User-Agent
        if isinstance(user_agent, str):
            request.headers['User-Agent'] = user_agent
        elif not user_agent:
            del request.headers['User-Agent']

        # Block when requesting the same URL multiple times simultaneously so the
        # first response can be loaded from cache by the other requests
        request_lock_key = (request.url, await request.aread())
        # _log.debug('Request lock key: %r', request_lock_key)
        request_lock = _request_locks[request_lock_key]
        async with request_lock:
            if cache:
                cache_file = _cache_file(method, url, params)
                result = _from_cache(cache_file, max_age=max_cache_age)
                if result is not None:
                    return result

            if debug_file:
                await _dump_req_res(request, f'{debug_file}.request')

            try:
                response = await client.send(
                    request=request,
                    auth=auth,
                    follow_redirects=follow_redirects,
                )
                try:
                    response.raise_for_status()
                except httpx.HTTPStatusError:
                    if response.status_code not in (301, 302, 303, 307, 308):
                        raise errors.RequestError(
                            f'{url}: {utils.html.as_text(response.text)}',
                            url=url,
                            text=response.text,
                            headers=response.headers,
                            status_code=response.status_code,
                        )

            except httpx.TimeoutException:
                raise errors.RequestError(f'{url}: Timeout')

            except ssl.SSLError as e:
                if hasattr(e, 'verify_message'):
                    # E.g. "certificate has expired"
                    msg = e.verify_message.capitalize()
                elif hasattr(e, 'reason'):
                    # E.g. "CERTIFICATE_VERIFY_FAILED"
                    msg = e.reason
                elif e.strerror:
                    # `strerror` is inherited from OSError and should be None if
                    # not provided
                    msg = e.strerror
                else:
                    msg = str(e)
                raise errors.RequestError(f'{url}: TLS: {msg}')

            except httpx.HTTPError as e:
                _log.debug(f'Unexpected HTTP error: {e!r}')
                raise errors.RequestError(f'{url}: {e}')

            else:
                _save_session_cookies(cookies=client.cookies, domain=response.url.host)
                if cookies and isinstance(cookies, (str, pathlib.Path)):
                    _save_permanent_cookies(client=client, filepath=cookies, domain=response.url.host)

                if cache:
                    cache_file = _cache_file(method, url, params)
                    _to_cache(cache_file, response.content)

                if debug_file:
                    await _dump_req_res(response, f'{debug_file}.response')

                return Response(
                    text=response.text,
                    bytes=response.content,
                    headers=response.headers,
                    status_code=response.status_code,
                )


async def _dump_req_res(r, filepath):
    content = await r.aread()

    if isinstance(r, httpx.Request):
        _log.debug('Sending request: %r', r)
        _log.debug('Request headers: %r', r.headers)
        _log.debug('Request data: %r', content)
    else:
        _log.debug('Received response: %r', r)
        _log.debug('Response headers: %r', r.headers)
        _log.debug('Response data: %r', content)

    utils.fs.mkdir(os.path.dirname(filepath))
    if isinstance(content, str):
        utils.html.dump(content, f'{filepath}.html')
    elif isinstance(content, bytes):
        with open(f'{filepath}.content', 'wb') as f:
            f.write(content)
    elif isinstance(content, str):
        with open(f'{filepath}.content', 'w') as f:
            f.write(content)
    else:
        raise RuntimeError(f'How should I dump {type(content).__name__}?: {content!r}')

    with open(f'{filepath}.headers', 'w') as f:
        f.write(json.dumps(dict(r.headers), indent=4))


def _load_session_cookies(domain):
    # _log.debug('Loading session cookies for %r: %r', domain, _session_cookies[domain])
    return _session_cookies[domain]


def _save_session_cookies(cookies, domain):
    _session_cookies[domain].update(cookies)
    # _log.debug('Saved session cookies for %r: %r: %r', domain, cookies, _session_cookies[domain])


def clear_session_cookies(domain=None):
    """
    Delete all session cookies for `domain`

    If `domain` is `None`, delete all session cookies.
    """
    if domain:
        _session_cookies[domain].clear()
    else:
        for domain in _session_cookies:
            _session_cookies[domain].clear()


def _load_permanent_cookies(cookies):
    if isinstance(cookies, (collections.abc.Mapping, http.cookiejar.CookieJar)):
        _log.debug('Loading permanent cookies from mapping: %r', cookies)
        return cookies
    elif cookies and isinstance(cookies, (str, pathlib.Path)):
        filepath_abs = os.path.join(_get_cache_directory(), cookies)
        _log.debug('Loading permanent cookies from file: %r', filepath_abs)
        cookie_jar = http.cookiejar.LWPCookieJar(filepath_abs)
        try:
            cookie_jar.load()
        except FileNotFoundError:
            _log.debug('No such file: %r', filepath_abs)
            pass
        except OSError as e:
            msg = e.strerror if e.strerror else str(e)
            raise errors.RequestError(f'Failed to read {cookie_jar.filename}: {msg}')
        else:
            _log.debug('Loaded permanent cookies: %r', cookie_jar)
        return cookie_jar
    elif cookies is not None:
        raise RuntimeError(f'Unsupported cookies type: {cookies!r}')


def _save_permanent_cookies(client, filepath, domain):
    filepath_abs = os.path.join(_get_cache_directory(), filepath)
    _log.debug('Saving permanent cookies for %r to %r', domain, filepath_abs)
    file_cookie_jar = http.cookiejar.LWPCookieJar(filepath_abs)
    for cookie in client.cookies.jar:
        if domain.endswith(cookie.domain):
            _log.debug('Saving permanent cookie: %r', cookie)
            file_cookie_jar.set_cookie(cookie)
        else:
            _log.debug('Ignoring cookie because %r does not end with %r: %r', domain, cookie.domain, cookie)

    if file_cookie_jar:
        try:
            utils.fs.mkdir(utils.fs.dirname(filepath_abs))
            file_cookie_jar.save()
        except OSError as e:
            msg = e.strerror if e.strerror else str(e)
            raise errors.RequestError(f'Failed to write {filepath_abs}: {msg}')
        else:
            _log.debug('Saved permanent cookies: %r', file_cookie_jar)


def _open_files(files):
    """
    Open files for upload

    See :func:`post` and
    https://www.python-httpx.org/advanced/#multipart-file-encoding.

    :raise RequestError: if a file cannot be opened
    """
    opened = {}
    for fieldname, fileinfo in files.items():
        # Field name is mapped to file path
        if isinstance(fileinfo, str):
            filename = os.path.basename(fileinfo)
            fileobj = _get_file_object(fileinfo)
            opened[fieldname] = (filename, fileobj)

        # Field name is mapped to opened file/file-like object/stream
        elif isinstance(fileinfo, io.BytesIO):
            opened[fieldname] = (None, fileinfo)

        # Field name is mapped to dictionary with the keys "file", "filename"
        # and "mimetype".
        # - "file" can be a path or file-like object (like above)
        # - "filename" is the name of the file that is passed to the server or
        #   `None`.
        # - "mimetype" is the MIME type, which is autodetected unless "filename"
        #   is `None`.
        elif isinstance(fileinfo, collections.abc.Mapping):
            if isinstance(fileinfo['file'], str):
                filename = fileinfo.get('filename', os.path.basename(fileinfo['file']))
                fileobj = _get_file_object(fileinfo['file'])
            elif isinstance(fileinfo['file'], io.BytesIO):
                filename = fileinfo.get('filename', None)
                fileobj = fileinfo['file']
            else:
                raise RuntimeError(f'Invalid "file" value in fileinfo: {fileinfo["file"]!r}')
            mimetype = fileinfo.get('mimetype', False)
            if mimetype is False:
                opened[fieldname] = (filename, fileobj)
            else:
                opened[fieldname] = (filename, fileobj, mimetype)

        else:
            raise RuntimeError(f'Invalid fileinfo: {fileinfo}')

    return opened

def _get_file_object(filepath):
    """
    Open `filepath` and return the file object

    :raise RequestError: if OSError is raised by `open`
    """
    try:
        return open(filepath, 'rb')
    except OSError as e:
        msg = e.strerror if e.strerror else 'Failed to open'
        raise errors.RequestError(f'{filepath}: {msg}')


def _to_cache(cache_file, bytes):
    if not isinstance(bytes, builtins.bytes):
        raise TypeError(f'Not a bytes object: {bytes!r}')

    # Try to remove <script> tags
    if b'<script' in bytes:
        try:
            string = str(bytes, encoding='utf-8', errors='strict')
        except UnicodeDecodeError:
            pass
        else:
            string = utils.html.purge_javascript(string)
            bytes = string.encode('utf-8')

    try:
        utils.fs.mkdir(utils.fs.dirname(cache_file))
        with open(cache_file, 'wb') as f:
            f.write(bytes)
    except OSError as e:
        raise RuntimeError(f'Unable to write cache file {cache_file}: {e}')


def _from_cache(cache_file, max_age=float('inf')):
    try:
        cache_mtime = os.stat(cache_file).st_mtime
    except FileNotFoundError:
        return None
    else:
        cache_age = time.time() - cache_mtime
        if cache_age <= max_age:
            bytes = _read_bytes_from_file(cache_file)
            text = _read_string_from_file(cache_file)
            if bytes and text:
                return Response(text=text, bytes=bytes)


def _read_string_from_file(filepath):
    bytes = _read_bytes_from_file(filepath)
    if bytes:
        return str(bytes, encoding='utf-8', errors='replace')


def _read_bytes_from_file(filepath):
    try:
        with open(filepath, 'rb') as f:
            return f.read()
    except OSError:
        pass


def _cache_file(method, url, params={}):
    def make_filename(method, url, params_str):
        if params_str:
            filename = f'{method.upper()}.{url}?{params_str}'
        else:
            filename = f'{method.upper()}.{url}'
        return filename.replace(' ', '+')

    if params:
        params_str = '&'.join((f'{k}={v}' for k,v in params.items()))
        if len(make_filename(method, url, params_str)) > 250:
            params_str = f'[HASH:{utils.semantic_hash(params)}]'
    else:
        params_str = ''

    return os.path.join(
        _get_cache_directory(),
        utils.fs.sanitize_filename(make_filename(method, url, params_str)),
    )


DEFAULT_USER_AGENT = 'Mozilla/5.0 (X11; Linux) Gecko/20100101 Firefox'
USER_AGENT_CACHE_MAX_AGE = 14 * 24 * 60 * 60
USER_AGENT_SOURCE_URLS = [
    'https://www.whatismybrowser.com/guides/the-latest-user-agent/firefox',
]

async def get_popular_user_agent():
    """
    Return commonly used user agent

    The most recent user agent is cached for `USER_AGENT_CACHE_MAX_AGE` seconds.

    If no user agent can be acquired, return `DEFAULT_USER_AGENT`.
    """
    user_agent = _get_user_agent_from_cache()
    if user_agent:
        return user_agent

    for url in USER_AGENT_SOURCE_URLS:
        user_agent = await _get_user_agent_from_url(url)
        if user_agent:
            _log.debug('Got fresh user agent: %r', user_agent)
            _cache_user_agent(user_agent)
            return user_agent

    _log.debug('Defaulting to hardcoded user agent: %s', DEFAULT_USER_AGENT)
    return DEFAULT_USER_AGENT

def _get_user_agent_from_cache():
    cache_filepath = _get_user_agent_cache_filepath()
    if os.path.exists(cache_filepath):
        cache_created = os.stat(cache_filepath).st_mtime
        cache_age = time.time() - cache_created
        if cache_age < USER_AGENT_CACHE_MAX_AGE:
            try:
                return open(cache_filepath, 'r').read().strip()
            except OSError as e:
                _log.debug('Failed to read user agent from %s: %r', cache_filepath, e)

async def _get_user_agent_from_url(url):
    try:
        response = await get(url, cache=False, user_agent=False)
    except errors.RequestError as e:
        _log.debug('Failed to fetch user agent from %s: %r', url, e)
    else:
        user_agents = re.findall(
            r'>\s*(Mozilla/5\.0\s*.*?Linux.*?\s*Gecko/\d+\s*.*\s*Firefox/[\d\.]+)\s*<',
            response,
        )
        if user_agents:
            _log.debug('Got user agents from %s:', url)
            for ua in user_agents:
                _log.debug(' * %r', ua)
            return random.choice(user_agents)

def _cache_user_agent(user_agent):
    cache_filepath = _get_user_agent_cache_filepath()
    try:
        with open(cache_filepath, 'w') as f:
            f.write(user_agent)
    except OSError as e:
        _log.debug('Failed to write user agent to %s: %r', cache_filepath, e)

def _get_user_agent_cache_filepath():
    return os.path.join(_get_cache_directory(), 'user-agent')



import abc
import functools

from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout.containers import ConditionalContainer, HSplit, Window, to_container
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.layout import walk

from .. import widgets

import logging  # isort:skip
_log = logging.getLogger(__name__)


class JobWidgetBase(abc.ABC):
    """User-interaction and information display for :class:`~.jobs.JobBase` instance"""

    _no_runtime_widget = Window(
        dont_extend_height=True,
        style='class:info',
    )

    def __init__(self, job, app):
        self._job = job
        self._app = app
        self.setup()
        self.job.signal.register('info', lambda _: self.invalidate())
        self.job.signal.register('warning', lambda _: self.invalidate())
        main_widget = HSplit(
            children=[
                # Status information or user interaction
                ConditionalContainer(
                    filter=Condition(lambda: not self.job.is_finished),
                    content=self._runtime_widget or self._no_runtime_widget,
                ),
                # Result
                ConditionalContainer(
                    filter=Condition(lambda: self.job.output),
                    content=self.output_widget,
                ),
                # Additional info that isn't part of the job's main result
                # (e.g. CreateTorrentJobWidget can show the files in the
                # torrent, but the output is the torrent file path.)
                ConditionalContainer(
                    filter=Condition(lambda: not self.job.is_finished and bool(self.job.info)),
                    content=self.info_widget,
                ),
                # Warnings
                ConditionalContainer(
                    filter=Condition(lambda: bool(self.job.warnings)),
                    content=self.warnings_widget,
                ),
                # Errors
                ConditionalContainer(
                    filter=Condition(lambda: bool(self.job.errors)),
                    content=self.errors_widget,
                ),
            ],
        )
        label = widgets.HLabel(
            group='jobs',
            text=self.job.label,
            style='class:label',
            content=main_widget,
        )
        self._container = ConditionalContainer(
            filter=Condition(lambda: self.job.errors or not self.job.hidden),
            content=label,
        )

    @property
    def job(self):
        """Underlying :class:`~.JobBase` instance"""
        return self._job

    @abc.abstractmethod
    def setup(self):
        """
        Called on object creation

        Create widgets and register :attr:`job` callbacks.
        """

    @property
    @abc.abstractmethod
    def runtime_widget(self):
        """
        Interactive or status that is displayed while this job is running

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object or
            `None`
        """

    @functools.cached_property
    def _runtime_widget(self):
        runtime_widget = self.runtime_widget
        if runtime_widget:
            container = to_container(runtime_widget)
            container.key_bindings = self.keybindings_local
            return container

    @property
    def output_widget(self):
        """
        Job :attr:`~.JobBase.output`

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object
        """
        return Window(
            style='class:output',
            # FIXME: If output is empty, prompt-toolkit ignores the
            #        "dont_extend_height" argument. Using a space (0x20) as a
            #        placeholder seems to prevent this issue.
            content=FormattedTextControl(lambda: '\n'.join(self.job.output) or ' '),
            dont_extend_height=True,
            wrap_lines=True,
        )

    @property
    def info_widget(self):
        """
        :attr:`~.JobBase.info` that is only displayed while this job is running

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object
        """
        return Window(
            style='class:info',
            content=FormattedTextControl(lambda: str(self.job.info)),
            dont_extend_height=True,
            wrap_lines=True,
        )

    @property
    def warnings_widget(self):
        """
        Any :attr:`~.JobBase.warnings`

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object
        """
        return Window(
            style='class:warning',
            content=FormattedTextControl(lambda: '\n'.join(str(e) for e in self.job.warnings)),
            dont_extend_height=True,
            wrap_lines=True,
        )

    @property
    def errors_widget(self):
        """
        Any :attr:`~.JobBase.errors`

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object
        """
        return Window(
            style='class:error',
            content=FormattedTextControl(lambda: '\n'.join(str(e) for e in self.job.errors)),
            dont_extend_height=True,
            wrap_lines=True,
        )

    @property
    def is_interactive(self):
        """Whether this job needs user interaction"""
        if self.runtime_widget:
            for c in walk(to_container(self.runtime_widget), skip_hidden=True):
                if isinstance(c, Window) and c.content.is_focusable():
                    return True
        return False

    @functools.cached_property
    def keybindings_global(self):
        """
        Application-wide :class:`prompt_toolkit.key_binding.KeyBindings` instance

        These keybindings are always active as long as the TUI is running,
        regardless of which widget is focused. This also makes it possible to
        bind keys for unfocusable wigets.

        To keep things tidy and intuitive, all keybindings should be sequences
        of multiple keys. The first key should always start with ``c-<x>`` where
        ``c`` stands for ``Control`` and ``<x>`` is a character that is
        associated with the job's :attr:`~.JobBase.label`.

        For example, :class:`~.CreateTorrentJob` keybindings start with ``c-t``
        and :class:`~.SceneCheckJob` keybindings start with ``c-s``.
        """
        return self._app.key_bindings

    @functools.cached_property
    def keybindings_local(self):
        """
        Widget-wide :class:`prompt_toolkit.key_binding.KeyBindings` instance

        These keybindings are only active as long as this widget is focused. If
        there are conflicts, local keybindings take precedence over
        :attr:`global keybindings <keybindings_global>`.
        """
        return KeyBindings()

    def invalidate(self):
        """Schedule redrawing of the TUI"""
        self._app.invalidate()

    def __pt_container__(self):
        return self._container



import argparse
import pathlib
import re
from unittest.mock import call, patch

import pytest

from upsies import defaults, errors, trackers, utils
from upsies.utils import argtypes


@pytest.mark.parametrize(
    argnames='argtype, values, exp_result',
    argvalues=(
        (int, '\t1, 3 ,,5,  ', [1, 3, 5]),
        (float, '1.2,3.4,5.9', [1.2, 3.4, 5.9]),
        (int, '1, foo, 5', argparse.ArgumentTypeError('Invalid value: foo')),
    ),
    ids=lambda v: repr(v),
)
def test_comma_separated(argtype, values, exp_result):
    func = argtypes.comma_separated(argtype)
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            func(values)
    else:
        return_value = func(values)
        assert return_value == exp_result


def test_client_valid_value(mocker):
    mocker.patch('upsies.utils.btclient.client_names', return_value=('foo', 'bar', 'baz'))
    assert argtypes.client('foo') == 'foo'
    assert argtypes.client('Bar') == 'bar'
    assert argtypes.client('BAZ') == 'baz'

def test_client_invalid_value(mocker):
    mocker.patch('upsies.utils.btclient.client_names', return_value=('foo', 'bar', 'baz'))
    with pytest.raises(argparse.ArgumentTypeError, match=r'^Unsupported client: asdf$'):
        argtypes.client('asdf')


def test_content(mocker):
    release_mock = mocker.patch('upsies.utils.argtypes.release', return_value='release/path')
    existing_path_mock = mocker.patch('upsies.utils.argtypes.existing_path', return_value='existing/path')
    assert argtypes.content('given/path') == 'existing/path'
    assert release_mock.call_args_list == [call('given/path')]
    assert existing_path_mock.call_args_list == [call('release/path')]


@pytest.mark.parametrize(
    argnames='path, path_exists, exp_exc',
    argvalues=(
        ('path/to/foo', True, None),
        (pathlib.Path('path/to/foo'), True, None),
        ('path/to/foo', False, argparse.ArgumentTypeError('No such file or directory: path/to/foo')),
    ),
    ids=lambda v: repr(v),
)
def test_existing_path(path, path_exists, exp_exc, mocker):
    exists_mock = mocker.patch('os.path.exists', return_value=path_exists)
    if path_exists:
        assert argtypes.existing_path(path) == str(path)
    else:
        with pytest.raises(type(exp_exc), match=rf'^{re.escape(str(exp_exc))}$'):
            argtypes.content(path)
    assert exists_mock.call_args_list == [call(str(path))]


@pytest.mark.parametrize('imghost', utils.imghosts.imghost_names())
def test_imghost_valid_value(imghost):
    assert argtypes.imghost(imghost) == imghost
    assert argtypes.imghost(imghost.upper()) == imghost
    assert argtypes.imghost(imghost.capitalize()) == imghost

def test_imghost_invalid_value():
    with pytest.raises(argparse.ArgumentTypeError, match=r'^Unsupported image hosting service: foo$'):
        argtypes.imghost('foo')


@pytest.mark.parametrize('imghost', utils.imghosts.imghost_names())
def test_imghosts_only_valid_values(imghost):
    existing_names = utils.imghosts.imghost_names()
    existing_names.insert(0, '')
    existing_names.insert(2, '')
    existing_names.append('')
    existing_names[2] = existing_names[2].upper()
    existing_names_combined = ','.join(existing_names)
    assert argtypes.imghosts(existing_names_combined) == utils.imghosts.imghost_names()

def test_imghosts_invalid_value():
    existing_names = utils.imghosts.imghost_names()
    existing_names.insert(2, 'foo')
    existing_names_combined = ','.join(existing_names)
    with pytest.raises(argparse.ArgumentTypeError, match=r'^Unsupported image hosting service: foo$'):
        argtypes.imghosts(existing_names_combined)


@pytest.mark.parametrize(
    argnames='value, exp_value',
    argvalues=(
        (123, 123),
        (123.2, 123),
        ('123', 123),
        ('123.2', 123),
    ),
    ids=lambda v: repr(v),
)
def test_integer_valid_value(value, exp_value):
    assert argtypes.integer(value) == exp_value

@pytest.mark.parametrize('value', ('one', (1, 2, 3)), ids=lambda v: repr(v))
def test_integer_invalid_value(value):
    with pytest.raises(argparse.ArgumentTypeError, match=rf'^Not an integer: {re.escape(repr(value))}'):
        argtypes.integer(value)


@pytest.mark.parametrize(
    argnames='min, max, value, exp_value',
    argvalues=(
        (1, 5, 5.0, 5),
    ),
)
def test_number_of_screenshots_valid_value(min, max, value, exp_value):
    assert argtypes.number_of_screenshots(min=min, max=max)(value) == exp_value

@pytest.mark.parametrize(
    argnames='min, max, value, exp_error',
    argvalues=(
        (1, 5, 0, 'Minimum is 1'),
        (1, 5, 6, 'Maximum is 5'),
        (1, 5, 'asdf', "Invalid integer value: 'asdf'"),
    ),
)
def test_number_of_screenshots_invalid_value(min, max, value, exp_error):
    with pytest.raises(argparse.ArgumentTypeError, match=rf'^{re.escape(exp_error)}$'):
        argtypes.number_of_screenshots(min=min, max=max)(value)


@pytest.mark.parametrize('option', defaults.option_paths())
def test_option_valid_value(option):
    assert argtypes.option(option) == option

def test_option_invalid_value():
    with pytest.raises(argparse.ArgumentTypeError, match=r'^Unknown option: foo$'):
        argtypes.option('foo')


@pytest.mark.parametrize(
    argnames='values, value, exp_result',
    argvalues=(
        (('foo', 'bar', 'baz'), 'foo', 'foo'),
        (('foo', 'bar', 'baz'), 'bar', 'bar'),
        (('foo', 'bar', 'baz'), 'baz', 'baz'),
        (('foo', 'bar', 'baz'), 'hello', argparse.ArgumentTypeError('Invalid value: hello')),
    ),
    ids=lambda v: repr(v),
)
def test_one_of(values, value, exp_result):
    func = argtypes.one_of(values)
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            func(value)
    else:
        assert func(value) == exp_result


@pytest.mark.parametrize(
    argnames='value, exp_regex',
    argvalues=(
        ('foo.*bar', re.compile(r'foo.*bar')),
    ),
)
def test_regex_valid_value(value, exp_regex):
    assert argtypes.regex(value) == exp_regex

@pytest.mark.parametrize(
    argnames='regex, exp_msg',
    argvalues=(
        ('foo[bar', 'unterminated character set at position 3'),
    ),
)
def test_regex_invalid_value(regex, exp_msg):
    msg = f'Invalid regular expression: {regex}: {exp_msg}'
    with pytest.raises(argparse.ArgumentTypeError, match=rf'^{re.escape(msg)}$'):
        argtypes.regex(regex)


@pytest.mark.parametrize(
    argnames='path, exp_path, exc, exp_exc',
    argvalues=(
        ('path/to/foo', 'path/to/foo', None, None),
        (123, '123', None, None),
        ('path/to/foo', 'path/to/foo', errors.SceneAbbreviatedFilenameError('bar'),
         argparse.ArgumentTypeError(errors.SceneAbbreviatedFilenameError('bar'))),
        ('path/to/foo', 'path/to/foo', RuntimeError('baz'), RuntimeError('baz')),
    ),
    ids=lambda v: str(v),
)
def test_release(path, exp_path, exc, exp_exc, mocker):
    mocker.patch('upsies.utils.predbs.assert_not_abbreviated_filename',
                 side_effect=exc)
    if exp_exc:
        with pytest.raises(type(exp_exc), match=rf'^{re.escape(str(exp_exc))}$'):
            argtypes.release(path)
    else:
        assert argtypes.release(path) == exp_path


@pytest.mark.parametrize('predb_name', utils.predbs.predb_names())
def test_predb_name_valid_value(predb_name):
    assert argtypes.predb_name(predb_name) == predb_name
    assert argtypes.predb_name(predb_name.upper()) == predb_name
    assert argtypes.predb_name(predb_name.capitalize()) == predb_name

def test_predb_name_invalid_value():
    with pytest.raises(argparse.ArgumentTypeError, match=r'^Unsupported scene release database: foo$'):
        argtypes.predb_name('foo')


@pytest.mark.parametrize('predb', utils.predbs.predb_names())
def test_predb_valid_value(predb):
    for return_value in (
            argtypes.predb(predb),
            argtypes.predb(predb.upper()),
            argtypes.predb(predb.capitalize()),
    ):
        assert isinstance(return_value, utils.predbs.PredbApiBase)
        assert return_value.name == predb.lower()

def test_predb_invalid_value():
    with pytest.raises(argparse.ArgumentTypeError, match=r'^Unsupported scene release database: foo$'):
        argtypes.predb('foo')


@pytest.mark.parametrize(
    argnames='value, exp_result',
    argvalues=(
        (None, None),
        ('true', True),
        ('false', False),
        ('maybe', argparse.ArgumentTypeError("Invalid boolean value: 'maybe'")),
    ),
)
def test_bool_or_none(value, exp_result):
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            argtypes.bool_or_none(value)
    else:
        return_value = argtypes.bool_or_none(value)
        assert return_value == exp_result


@patch('upsies.utils.timestamp.parse')
def test_timestamp_valid_value(parse_mock):
    assert argtypes.timestamp('foo') is parse_mock.return_value
    assert parse_mock.call_args_list == [call('foo')]

@patch('upsies.utils.timestamp.parse')
@pytest.mark.parametrize('exception', (ValueError, TypeError))
def test_timestamp_invalid_value(parse_mock, exception):
    parse_mock.side_effect = exception('Message!')
    with pytest.raises(argparse.ArgumentTypeError, match=r'^Message!$'):
        argtypes.timestamp('foo')


@pytest.mark.parametrize('tracker', trackers.tracker_names())
def test_tracker_valid_value(tracker):
    assert argtypes.tracker(tracker) == tracker
    assert argtypes.tracker(tracker.upper()) == tracker
    assert argtypes.tracker(tracker.capitalize()) == tracker

def test_tracker_invalid_value():
    with pytest.raises(argparse.ArgumentTypeError, match=r'^Unsupported tracker: foo$'):
        argtypes.tracker('foo')


@pytest.mark.parametrize('webdb', utils.webdbs.webdb_names())
def test_webdb_valid_value(webdb):
    assert argtypes.webdb(webdb) == webdb
    assert argtypes.webdb(webdb.upper()) == webdb
    assert argtypes.webdb(webdb.capitalize()) == webdb

def test_webdb_invalid_value():
    with pytest.raises(argparse.ArgumentTypeError, match=r'^Unsupported database: foo$'):
        argtypes.webdb('foo')



"""
Create screenshots from video file and optionally upload them
"""

import functools

from .... import jobs, utils
from .base import CommandBase


class screenshots(CommandBase):
    """Create screenshots from video file and optionally upload them"""

    names = ('screenshots', 'ss')

    argument_definitions = {
        'CONTENT': {
            'type': utils.argtypes.content,
            'help': 'Path to release content',
        },
        ('--exclude-files', '--ef'): {
            'nargs': '+',
            'metavar': 'PATTERN',
            'help': ('Glob pattern to exclude from CONTENT '
                     '(matched case-insensitively against path relative to CONTENT)'),
            'default': (),
        },
        ('--exclude-files-regex', '--efr'): {
            'nargs': '+',
            'metavar': 'PATTERN',
            'help': ('Regular expression to exclude from CONTENT '
                     '(matched case-sensitively against path relative to CONTENT)'),
            'type': utils.argtypes.regex,
            'default': (),
        },
        ('--timestamps', '-t'): {
            'nargs': '+',
            'default': (),
            'type': utils.argtypes.timestamp,
            'metavar': 'TIMESTAMP',
            'help': 'Space-separated list of [[HH:]MM:]SS strings',
        },
        ('--number', '-n'): {
            'type': utils.argtypes.integer,
            'help': 'How many screenshots to make in total',
            'default': 0,
        },
        ('--from-all-videos', '-a'): {
            'action': 'store_true',
            'help': 'Make NUMBER screenshots from each video file beneath CONTENT',
        },
        ('--optimize', '--opt'): {
            'type': utils.argtypes.one_of(utils.image.optimization_levels),
            'default': None,
            'metavar': 'LEVEL',
            'help': f'File size optimization level: {", ".join(utils.image.optimization_levels)}',
        },
        ('--upload-to', '-u'): {
            'type': utils.argtypes.imghosts,
            'metavar': 'IMAGE_HOSTS',
            'help': (
                'Comma-separated list of case-insensitive image hosting service names\n'
                'Supported services: ' + ', '.join(utils.imghosts.imghost_names())
            ),
        },
        ('--output-directory', '-o'): {
            'default': '',  # Current working directory
            'metavar': 'PATH',
            'help': 'Directory where screenshots are put (created on demand)',
        },
    }

    @functools.cached_property
    def screenshots_job(self):
        return jobs.screenshots.ScreenshotsJob(
            home_directory=self.args.output_directory,
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
            content_path=self.args.CONTENT,
            exclude_files=(
                tuple(self.args.exclude_files)
                + tuple(self.args.exclude_files_regex)
            ),
            timestamps=self.args.timestamps,
            count=self.args.number,
            from_all_videos=self.args.from_all_videos,
            optimize=(
                self.args.optimize
                if self.args.optimize is not None else
                self.config['config']['screenshots']['optimize']
            ),
        )

    @functools.cached_property
    def upload_screenshots_job(self):
        if self.args.upload_to:
            imghost_job = jobs.imghost.ImageHostJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                imghosts=self.imghosts,
            )
            # Timestamps and number of screenshots are determined in a
            # subprocess, we have to wait for that before we can set the number
            # of expected screenhots.
            self.screenshots_job.signal.register('screenshots_total', imghost_job.set_images_total)
            # Pass ScreenshotsJob's output to ImageHostJob input.
            self.screenshots_job.signal.register('output', imghost_job.enqueue)
            # Tell imghost_job to finish the current upload and then finish.
            self.screenshots_job.signal.register('finished', lambda _: imghost_job.finalize())
            return imghost_job

    @functools.cached_property
    def imghosts(self):
        return tuple(
            utils.imghosts.imghost(
                name=name,
                options=self.config['imghosts'][name],
            )
            for name in self.args.upload_to
        )

    @functools.cached_property
    def jobs(self):
        return (
            self.screenshots_job,
            self.upload_screenshots_job,
        )
